from __future__ import annotations

import json
import sys

import iotbx.phil

import dials.util
from dials.algorithms.spot_finding import per_image_analysis
from dials.util import tabulate
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files

help_message = """

Reports the number of strong spots and computes an estimate of the resolution
limit for each image, given the results of dials.find_spots. Optionally
generates a plot of the per-image statistics (plot=image.png).

Examples::

  dials.spot_counts_per_image imported.expt strong.refl

  dials.spot_counts_per_image imported.expt strong.refl plot=per_image.png
"""

phil_scope = iotbx.phil.parse(
    """\
resolution_analysis = True
  .type = bool
plot = None
  .type = path
json = None
  .type = path
split_json = False
  .type = bool
joint_json = True
  .type = bool
id = None
  .type = int(value_min=0)
"""
)


@dials.util.show_mail_handle_errors()
def run(args=None):
    usage = "dials.spot_counts_per_image [options] imported.expt strong.refl"

    parser = ArgumentParser(
        usage=usage,
        read_reflections=True,
        read_experiments=True,
        phil=phil_scope,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=False)
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if not reflections and not experiments:
        parser.print_help()
        return

    # FIXME may want to change this to allow many to be passed i.e.
    # from parallel runs
    if len(reflections) != 1:
        sys.exit("Only one reflection list may be passed")
    reflections = reflections[0]

    if "miller_index" in reflections:
        sys.exit("Only unindexed reflections are currently supported")

    if any(experiments.crystals()):
        sys.exit("Only unindexed experiments are currently supported")

    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)

    if params.id is not None:
        reflections = reflections.select(reflections["id"] == params.id)

    all_stats = []
    for i, expt in enumerate(experiments):
        refl = reflections.select(reflections["id"] == i)
        stats = per_image_analysis.stats_per_image(
            expt, refl, resolution_analysis=params.resolution_analysis
        )
        all_stats.append(stats)

    # transpose stats
    summary_table = {}
    for s in all_stats:
        for k, value in s._asdict().items():
            summary_table.setdefault(k, [])
            summary_table[k].extend(value)
    stats = per_image_analysis.StatsMultiImage(**summary_table)
    print(stats)

    overall_stats = per_image_analysis.stats_for_reflection_table(
        reflections, resolution_analysis=params.resolution_analysis
    )
    rows = [
        ("Overall statistics", ""),
        ("#spots", "%i" % overall_stats.n_spots_total),
        ("#spots_no_ice", "%i" % overall_stats.n_spots_no_ice),
        ("d_min", f"{overall_stats.estimated_d_min:.2f}"),
        (
            "d_min (distl method 1)",
            "%.2f (%.2f)"
            % (overall_stats.d_min_distl_method_1, overall_stats.noisiness_method_1),
        ),
        (
            "d_min (distl method 2)",
            "%.2f (%.2f)"
            % (overall_stats.d_min_distl_method_2, overall_stats.noisiness_method_2),
        ),
    ]
    print(tabulate(rows, headers="firstrow"))

    if params.json:
        if params.split_json:
            for k, v in stats._asdict().items():
                start, end = params.json.split(".")
                with open(f"{start}_{k}.{end}", "w") as fp:
                    json.dump(v, fp)
        if params.joint_json:
            with open(params.json, "w") as fp:
                json.dump(stats._asdict(), fp)
    if params.plot:
        import matplotlib

        matplotlib.use("Agg")
        per_image_analysis.plot_stats(stats, filename=params.plot)


if __name__ == "__main__":
    run()



from __future__ import annotations

from dials_algorithms_image_connected_components_ext import *  # noqa: F403; lgtm

__all__ = ("LabelImageStack2d", "LabelImageStack3d", "LabelPixels3d")  # noqa: F405



from __future__ import annotations

import math

# constants removed from elsewhere - EPS which is the "standard small number"
EPS = 1e-7

# FULL_PARTIALITY which corresponds to the greatest partiality we can
# reasonably expect integrating to +/- 3 sigma
FULL_PARTIALITY = math.erf(3 / math.sqrt(2))



from __future__ import annotations

from collections import namedtuple

from scitbx import matrix

from dials.algorithms.refinement.parameterisation.detector_parameters import (
    DetectorMixin,
)
from dials.algorithms.refinement.parameterisation.scan_varying_model_parameters import (
    GaussianSmoother,
    ScanVaryingModelParameterisation,
    ScanVaryingParameterSet,
)

# bucket to provide value and axis attributes to model the Parameter class
Param = namedtuple("Param", ["value", "axis"])


class ScanVaryingDetectorParameterisationSinglePanel(
    ScanVaryingModelParameterisation, DetectorMixin
):
    """A scan-varying parameterisation for a single abstract panel plane, with
    angles expressed in mrad"""

    def __init__(self, detector, t_range, num_intervals, experiment_ids=None):

        if experiment_ids is None:
            experiment_ids = [0]

        # Set up the smoother
        smoother = GaussianSmoother(t_range, num_intervals)
        nv = smoother.num_values()

        # Factory function to provide to _init_core
        def parameter_type(value, axis, ptype, name):
            return ScanVaryingParameterSet(value, nv, axis, ptype, name)

        # Set up the initial state and parameter list
        dat = self._init_core(detector, parameter_type)

        self._d_at_t = matrix.sqr(detector[0].get_d_matrix())

        # set up the base class
        ScanVaryingModelParameterisation.__init__(
            self,
            detector,
            dat["istate"],
            dat["p_list"],
            smoother,
            experiment_ids=experiment_ids,
        )

        return

    def compose(self, t):
        """calculate state and derivatives for model at image number t"""

        # extract parameter sets from the internal list
        dist_set, shift1_set, shift2_set, tau1_set, tau2_set, tau3_set = self._param

        # extract data at time t using the smoother
        dist, dist_weights, dist_sumweights = self._smoother.value_weight(t, dist_set)
        shift1, shift1_weights, shift1_sumweights = self._smoother.value_weight(
            t, shift1_set
        )
        shift2, shift2_weights, shift2_sumweights = self._smoother.value_weight(
            t, shift2_set
        )
        tau1, tau1_weights, tau1_sumweights = self._smoother.value_weight(t, tau1_set)
        tau2, tau2_weights, tau2_sumweights = self._smoother.value_weight(t, tau2_set)
        tau3, tau3_weights, tau3_sumweights = self._smoother.value_weight(t, tau3_set)

        # calculate derivatives of values at image t wrt underlying parameters.
        ddist_dp = dist_weights * (1.0 / dist_sumweights)
        dshift1_dp = shift1_weights * (1.0 / shift1_sumweights)
        dshift2_dp = shift2_weights * (1.0 / shift2_sumweights)
        dtau1_dp = tau1_weights * (1.0 / tau1_sumweights)
        dtau2_dp = tau2_weights * (1.0 / tau2_sumweights)
        dtau3_dp = tau3_weights * (1.0 / tau3_sumweights)

        # calculate new [d] and derivatives wrt the values dist, shift1, shift2,
        # tau1, tau2 and tau3
        new_state, dd_dval = self._compose_core(
            Param(dist, dist_set.axis),
            Param(shift1, shift1_set.axis),
            Param(shift2, shift2_set.axis),
            Param(tau1, tau1_set.axis),
            Param(tau2, tau2_set.axis),
            Param(tau3, tau3_set.axis),
        )

        # 'fast' axis elements
        f1, f2, f3 = new_state["d1"]

        # 'slow' axis elements
        s1, s2, s3 = new_state["d2"]

        # origin vector elements
        o1, o2, o3 = new_state["origin"]

        self._d_at_t = matrix.sqr((f1, s1, o1, f2, s2, o2, f3, s3, o3))

        # calculate derivatives of state wrt underlying smoother parameters
        dd_dp1 = [None] * ddist_dp.size
        for (i, v) in ddist_dp:
            dd_dp1[i] = dd_dval[0] * v
        dd_dp2 = [None] * dshift1_dp.size
        for (i, v) in dshift1_dp:
            dd_dp2[i] = dd_dval[1] * v
        dd_dp3 = [None] * dshift2_dp.size
        for (i, v) in dshift2_dp:
            dd_dp3[i] = dd_dval[2] * v
        dd_dp4 = [None] * dtau1_dp.size
        for (i, v) in dtau1_dp:
            dd_dp4[i] = dd_dval[3] * v
        dd_dp5 = [None] * dtau2_dp.size
        for (i, v) in dtau2_dp:
            dd_dp5[i] = dd_dval[4] * v
        dd_dp6 = [None] * dtau3_dp.size
        for (i, v) in dtau3_dp:
            dd_dp6[i] = dd_dval[5] * v

        # store derivatives as list-of-lists
        self._dstate_dp = [dd_dp1, dd_dp2, dd_dp3, dd_dp4, dd_dp5, dd_dp6]

        return

    def get_state(self):
        """Return detector matrix [d] at image number t"""
        # only a single panel exists, so no multi_state_elt argument is allowed
        return self._d_at_t



from __future__ import annotations

import binascii
import os
import random

from dxtbx.ext import compress


def gz_open(filename, mode):
    import gzip

    return gzip.GzipFile(filename, mode)


def split_counts(image, split):
    from scitbx.array_family import flex

    new_images = [flex.int(flex.grid(image.focus()), 0) for k in range(split)]

    negative = image.as_1d() < 0
    positive = image.as_1d() > 0

    for new_image in new_images:
        new_image.as_1d().set_selected(negative, image.as_1d())

    for p in positive.iselection():
        counts = image[p]
        for j in range(counts):
            new_images[random.randint(0, split - 1)][p] += 1

    return new_images


def merge_counts(images):
    from scitbx.array_family import flex

    image = flex.int(flex.grid(images[0].focus()), 0)
    negative = images[0].as_1d() < 0
    for i in images:
        image += i
    image.as_1d().set_selected(negative, images[0].as_1d())
    return image


def read_image(in_image):
    from dxtbx import load

    assert os.path.exists(in_image)

    start_tag = binascii.unhexlify("0c1a04d5")

    data = gz_open(in_image, "rb").read()
    data_offset = data.find(start_tag)
    cbf_header = data[:data_offset]
    pixel_values = load(in_image).get_raw_data()

    return pixel_values, cbf_header


def write_image(out_image, pixel_values, header, nn=1):
    assert not os.path.exists(out_image)
    start_tag = binascii.unhexlify("0c1a04d5")

    compressed = compress(pixel_values)

    fixed_header = ""
    header = header.decode()
    for record in header.split("\n")[:-1]:
        if "X-Binary-Size:" in record:
            fixed_header += f"X-Binary-Size: {len(compressed)}\r\n"
        elif "Content-MD5" in record:
            pass
        elif "Count_cutoff" in record:
            cutoff = int(record.split()[2]) * nn
            fixed_header += "# Count_cutoff %d counts\n" % cutoff
        else:
            fixed_header += f"{record}\n"

    tailer = "\r\n--CIF-BINARY-FORMAT-SECTION----\r\n;\r\n"

    gz_open(out_image, "wb").write(
        fixed_header.encode() + start_tag + compressed + tailer.encode()
    )


def main(in_images, out_images):
    assert len(in_images) == len(out_images)
    n = len(in_images)

    for i in in_images:
        assert os.path.exists(i)
    for o in out_images:
        assert not os.path.exists(o)

    in_image_data = []
    in_image_headers = []

    for i in in_images:
        print(f"Reading {i}")
        pixel, header = read_image(i)
        in_image_data.append(pixel)
        in_image_headers.append(header)

    sum_image = merge_counts(in_image_data)
    rebin_images = split_counts(sum_image, n)

    for o, pixel, header in zip(out_images, rebin_images, in_image_headers):
        print(f"Writing {o}")
        write_image(o, pixel, header)


def main_sum(in_images, out_image):
    for i in in_images:
        assert os.path.exists(i)
    assert not os.path.exists(out_image)

    in_image_data = []
    in_image_headers = []

    for i in in_images:
        print(f"Reading {i}")
        pixel, header = read_image(i)
        in_image_data.append(pixel)
        in_image_headers.append(header)

    sum_image = merge_counts(in_image_data)

    print(f"Writing {out_image}")
    write_image(out_image, sum_image, in_image_headers[0], nn=len(in_images))



"""Tests for dials.report"""
from __future__ import annotations

import json

import procrunner


def test_report_integrated_data(dials_data, tmpdir):
    """Simple test to check that dials.report completes when given integrated data."""
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    result = procrunner.run(
        [
            "dials.report",
            data_dir / "20_integrated_experiments.json",
            data_dir / "20_integrated.pickle",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("dials.report.html").check()


def test_report_scaled_data(dials_data, tmpdir):
    """Test that dials.report works on scaled data."""
    data_dir = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    result = procrunner.run(
        [
            "dials.report",
            data_dir / "scaled_30.refl",
            data_dir / "scaled_30.expt",
            f"json={tmpdir}/report.json",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("dials.report.html").check()
    report_json = tmpdir.join("report.json")
    assert report_json.check()
    expected_keys = {
        "strong",
        "centroid",
        "intensity",
        "reference",
        "scan_varying",
        "scaling_model",
        "resolution_graphs",
        "batch_graphs",
        "misc_graphs",
        "scaled_intensity_graphs",
        "scaling_tables",
        "xtriage_output",
        "image_range_tables",
    }
    with report_json.open(encoding="utf-8") as fh:
        d = json.load(fh)
        assert not expected_keys - set(d.keys())



from __future__ import annotations

import logging
import sys

from libtbx.phil import parse

from dials.util import Sorry, log, show_mail_handle_errors

logger = logging.getLogger("dials.command_line.export_best")

help_message = """
This program is used to export the results of dials processing for the strategy
program BEST.
"""

phil_scope = parse(
    """

  n_bins = 100
    .type = int(value_min=1)
    .help = "Number of resolution bins for background estimation"

  min_partiality = 0.1
    .type = float(value_min=0, value_max=1)
    .help = "Minimum partiality of reflections to export"

  output {

    log = dials.export_best.log
      .type = path
      .help = "The log filename"

    prefix = best
      .type = str
      .help = "The prefix for the output file names for best"
              "(.hkl, .dat and .par files)"

  }
"""
)


class BestExporter:
    def __init__(self, params, experiments, reflections):
        """
        Initialise the exporter

        :param params: The phil parameters
        :param experiments: The experiment list
        :param reflections: The reflection tables
        """

        # Check the input
        if not experiments:
            raise Sorry("BEST exporter requires an experiment list")
        if not reflections:
            raise Sorry("BEST exporter require a reflection table")

        # Save the stuff
        self.params = params
        self.experiments = experiments
        self.reflections = reflections

    def export(self):
        """
        Export the files
        """
        from dials.util import best

        experiment = self.experiments[0]
        reflections = self.reflections[0]
        partiality = reflections["partiality"]
        sel = partiality >= self.params.min_partiality
        logger.info(
            "Selecting %s/%s reflections with partiality >= %s",
            sel.count(True),
            sel.size(),
            self.params.min_partiality,
        )
        if sel.count(True) == 0:
            raise Sorry(
                "No reflections remaining after filtering for minimum partiality (min_partiality=%f)"
                % (self.params.min_partiality)
            )
        reflections = reflections.select(sel)

        imageset = experiment.imageset
        prefix = self.params.output.prefix

        best.write_background_file(f"{prefix}.dat", imageset, n_bins=self.params.n_bins)
        best.write_integrated_hkl(prefix, reflections)
        best.write_par_file(f"{prefix}.par", experiment)


@show_mail_handle_errors()
def run(args=None):
    from dials.util.options import (
        ArgumentParser,
        reflections_and_experiments_from_files,
    )
    from dials.util.version import dials_version

    usage = "dials.export models.expt reflections.refl [options]"

    parser = ArgumentParser(
        usage=usage,
        read_experiments=True,
        read_reflections=True,
        phil=phil_scope,
        epilog=help_message,
    )

    # Get the parameters
    params, options = parser.parse_args(args, show_diff_phil=False)

    # Configure the logging
    log.config(logfile=params.output.log)

    # Print the version number
    logger.info(dials_version())

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    if not params.input.experiments and not params.input.reflections:
        parser.print_help()
        sys.exit()

    # Get the experiments and reflections
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    exporter = BestExporter(params, experiments, reflections)
    exporter.export()


if __name__ == "__main__":
    run()



from __future__ import annotations

import procrunner
import pytest


@pytest.mark.parametrize("filename", ["image_15799_master.h5", "image_15799.nxs"])
def test_convert_to_cbf(dials_data, filename, tmp_path):
    result = procrunner.run(
        [
            "dials.import",
            dials_data("vmxi_thaumatin", pathlib=True) / filename,
            "image_range=1,10",
        ],
        working_directory=tmp_path,
    )
    result.check_returncode()
    assert not result.stderr
    assert tmp_path.joinpath("imported.expt").is_file()

    result = procrunner.run(
        ["dials.convert_to_cbf", "imported.expt"], working_directory=tmp_path
    )
    result.check_returncode()
    assert not result.stderr

    assert len(list(tmp_path.glob("as_cbf_*.cbf"))) == 10



from __future__ import annotations


class BackgroundAlgorithm:
    """Class to do background subtraction."""

    def __init__(
        self, experiments, model="constant3d", tuning_constant=1.345, min_pixels=10
    ):
        """
        Initialise the algorithm.

        :param experiments: The list of experiments
        :param model: The background model
        :param tuning_constant: The robust tuning constant
        """
        from dials.algorithms.background.glm import Creator

        if model == "constant2d":
            model = Creator.model.constant2d
        elif model == "constant3d":
            model = Creator.model.constant3d
        elif model == "loglinear2d":
            model = Creator.model.loglinear2d
        elif model == "loglinear3d":
            model = Creator.model.loglinear3d
        else:
            raise RuntimeError("Unknown background model")
        self._create = Creator(
            model=model,
            tuning_constant=tuning_constant,
            max_iter=100,
            min_pixels=min_pixels,
        )

    def compute_background(self, reflections, image_volume=None):
        """
        Compute the background.

        :param reflections: The list of reflections
        """
        # Do the background subtraction
        if image_volume is None:
            success = self._create(reflections["shoebox"])
            reflections["background.mean"] = reflections[
                "shoebox"
            ].mean_modelled_background()
        else:
            success = self._create(reflections, image_volume)
        reflections.set_flags(~success, reflections.flags.dont_integrate)
        return success


class GLMBackgroundCalculatorFactory:
    """Class to do background subtraction."""

    @staticmethod
    def create(experiments, model="constant3d", tuning_constant=1.345, min_pixels=10):
        """
        Initialise the algorithm.

        :param experiments: The list of experiments
        :param model: The background model
        :param tuning_constant: The robust tuning constant
        """
        from dials.algorithms.background.glm import Creator
        from dials.algorithms.integration.parallel_integrator import (
            GLMBackgroundCalculator,
        )

        if model == "constant2d":
            model = Creator.model.constant2d
        elif model == "constant3d":
            model = Creator.model.constant3d
        elif model == "loglinear2d":
            model = Creator.model.loglinear2d
        elif model == "loglinear3d":
            model = Creator.model.loglinear3d
        else:
            raise RuntimeError("Unknown background model")
        return GLMBackgroundCalculator(
            model=model,
            tuning_constant=tuning_constant,
            max_iter=100,
            min_pixels=min_pixels,
        )



from __future__ import annotations

import functools

from orderedset import OrderedSet

from dxtbx.model.experiment_list import ExperimentList
from libtbx.phil import parse

import dials.util
from dials.array_family import flex
from dials.util import Sorry
from dials.util.export_mtz import match_wavelengths
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files

help_message = """

Utility script to split experiments and reflections from single files into
multiple files with one experiment per output experiment file and one
reflection file per output experiment file. Alternative options include
splitting by unique detector model or splitting by wavelength. The data
can also be split into chunks rather than individual output files if
required.

Example::

  dials.split_experiments combined.expt combined.refl
"""


class Script:
    def __init__(self):
        """Initialise the script."""
        # The phil scope
        phil_scope = parse(
            """
      by_detector = False
        .type = bool
        .help = "If True, instead of producing separate files for each"
                "experiment, experiments are grouped by unique detector"
                "model in the input set of experiments. For example, if"
                "there are five detector models in the input data, five"
                "sets of files will be produced, each containing"
                "experiments that reference a single detector model."
      by_wavelength = False
        .type = bool
        .help = "If True, group experiments by wavelength, from low to high"
                "(using a relative tolerance of 1e-4 to match wavelengths)."
      output {
        experiments_prefix = split
          .type = str
          .help = "Filename prefix for the split experimental models"

        reflections_prefix = split
          .type = str
          .help = "Filename prefix for the split reflections"

        template = "{prefix}_{index:0{maxindexlength:d}d}.{extension}"
          .type = str
          .expert_level = 2
          .help = "Template python format string for output filenames."
                  "Replaced variables are prefix (with"
                  "output.{experiments_prefix, reflections_prefix}),"
                  "index (number of split experiment), maxindexlength"
                  "(number of digits of total number of split experiments)"
                  "and extension (default file extension for model and"
                  "reflection files)"

        chunk_size = None
          .type = int
          .expert_level = 2
          .help = "If not None, instead of creating many individual"
                  "files, create composite files with no more than"
                  "chunk_size experiments per file."
        chunk_sizes = None
          .type = ints
          .expert_level = 2
          .help = "If not None, instead of creating many individual"
                  "files, create composite files with the number of"
                  "datasets given in the chunk_sizes list."

      }
    """,
            process_includes=True,
        )

        # The script usage
        usage = (
            "usage: dials.split_experiments [options] [param.phil] "
            "experiments1.expt experiments2.expt reflections1.refl "
            "reflections2.refl..."
        )

        # Create the parser
        self.parser = ArgumentParser(
            usage=usage,
            phil=phil_scope,
            read_reflections=True,
            read_experiments=True,
            check_format=False,
            epilog=help_message,
        )

    def run(self, args=None):
        """Execute the script."""

        # Parse the command line
        params, _ = self.parser.parse_args(args, show_diff_phil=True)

        # Try to load the models and data
        if not params.input.experiments:
            print("No Experiments found in the input")
            self.parser.print_help()
            return
        if params.input.reflections:
            if len(params.input.reflections) != len(params.input.experiments):
                raise Sorry(
                    "The number of input reflections files does not match the "
                    "number of input experiments"
                )
        reflections, experiments = reflections_and_experiments_from_files(
            params.input.reflections, params.input.experiments
        )
        if reflections:
            reflections = reflections[0]
        else:
            reflections = None

        experiments_template = functools.partial(
            params.output.template.format,
            prefix=params.output.experiments_prefix,
            maxindexlength=len(str(len(experiments) - 1)),
            extension="expt",
        )

        reflections_template = functools.partial(
            params.output.template.format,
            prefix=params.output.reflections_prefix,
            maxindexlength=len(str(len(experiments) - 1)),
            extension="refl",
        )

        if params.output.chunk_sizes:
            if not sum(params.output.chunk_sizes) == len(experiments):
                raise Sorry(
                    "Sum of chunk sizes list (%s) not equal to number of experiments (%s)"
                    % (sum(params.output.chunk_sizes), len(experiments))
                )

        if params.by_wavelength:
            if reflections:
                if not reflections.experiment_identifiers():
                    raise Sorry(
                        "Unable to split by wavelength as no experiment "
                        "identifiers are set in the reflection table."
                    )
            if all(experiments.identifiers() == ""):
                raise Sorry(
                    "Unable to split by wavelength as no experiment "
                    "identifiers are set in the experiment list."
                )

            wavelengths = match_wavelengths(experiments)
            for i, wl in enumerate(sorted(wavelengths.keys())):
                expids = []
                new_exps = ExperimentList()
                exp_nos = wavelengths[wl]
                imageset_ids = []  # record imageset ids to set in refl table
                imagesets_found = OrderedSet()
                for j in exp_nos:
                    expids.append(experiments[j].identifier)  # string
                    new_exps.append(experiments[j])
                    imagesets_found.add(experiments[j].imageset)
                    imageset_ids.append(imagesets_found.index(experiments[j].imageset))
                experiment_filename = experiments_template(index=i)
                print(
                    f"Saving experiments with wavelength {wl} to {experiment_filename}"
                )
                new_exps.as_json(experiment_filename)
                if reflections:
                    refls = reflections.select_on_experiment_identifiers(expids)
                    refls["imageset_id"] = flex.int(refls.size(), 0)
                    # now set the imageset ids
                    for k, iset_id in enumerate(imageset_ids):
                        # select the experiment based on id (unique per sweep),
                        # and set the imageset_id (not necessarily unique per sweep
                        # if imageset is shared)
                        sel = refls["id"] == k
                        refls["imageset_id"].set_selected(sel, iset_id)
                    reflections_filename = reflections_template(index=i)
                    print(
                        "Saving reflections with wavelength %s to %s"
                        % (wl, reflections_filename)
                    )
                    refls.as_file(reflections_filename)

        elif params.by_detector:
            assert (
                not params.output.chunk_size
            ), "chunk_size + by_detector is not implemented"
            if reflections is None:
                split_data = {
                    detector: {"experiments": ExperimentList()}
                    for detector in experiments.detectors()
                }
            else:
                split_data = {
                    detector: {
                        "experiments": ExperimentList(),
                        "reflections": flex.reflection_table(),
                        "imagesets_found": OrderedSet(),
                    }
                    for detector in experiments.detectors()
                }
            for i, experiment in enumerate(experiments):
                split_expt_id = experiments.detectors().index(experiment.detector)
                experiment_filename = experiments_template(index=split_expt_id)
                print("Adding experiment %d to %s" % (i, experiment_filename))
                split_data[experiment.detector]["experiments"].append(experiment)
                if reflections is not None:
                    reflections_filename = reflections_template(index=split_expt_id)
                    split_data[experiment.detector]["imagesets_found"].add(
                        experiment.imageset
                    )
                    print(
                        "Adding reflections for experiment %d to %s"
                        % (i, reflections_filename)
                    )
                    if reflections.experiment_identifiers().keys():
                        # first find which id value corresponds to experiment in question
                        identifier = experiment.identifier
                        id_ = None
                        for k in reflections.experiment_identifiers().keys():
                            if reflections.experiment_identifiers()[k] == identifier:
                                id_ = k
                                break
                        if id_ is None:
                            raise Sorry(
                                "Unable to find id matching experiment identifier in reflection table."
                            )
                        ref_sel = reflections.select(reflections["id"] == id_)
                        # now reset ids and reset/update identifiers map
                        for k in ref_sel.experiment_identifiers().keys():
                            del ref_sel.experiment_identifiers()[k]
                        new_id = len(split_data[experiment.detector]["experiments"]) - 1
                        ref_sel["id"] = flex.int(len(ref_sel), new_id)
                        ref_sel.experiment_identifiers()[new_id] = identifier
                    else:
                        ref_sel = reflections.select(reflections["id"] == i)
                        ref_sel["id"] = flex.int(
                            len(ref_sel),
                            len(split_data[experiment.detector]["experiments"]) - 1,
                        )
                    iset_id = split_data[experiment.detector]["imagesets_found"].index(
                        experiment.imageset
                    )
                    ref_sel["imageset_id"] = flex.int(ref_sel.size(), iset_id)
                    split_data[experiment.detector]["reflections"].extend(ref_sel)

            for i, detector in enumerate(experiments.detectors()):
                experiment_filename = experiments_template(index=i)
                print("Saving experiment %d to %s" % (i, experiment_filename))
                split_data[detector]["experiments"].as_json(experiment_filename)

                if reflections is not None:
                    reflections_filename = reflections_template(index=i)
                    print(
                        "Saving reflections for experiment %d to %s"
                        % (i, reflections_filename)
                    )
                    split_data[detector]["reflections"].as_file(reflections_filename)
        elif params.output.chunk_size or params.output.chunk_sizes:

            def save_chunk(chunk_id, expts, refls):
                experiment_filename = experiments_template(index=chunk_id)
                print("Saving chunk %d to %s" % (chunk_id, experiment_filename))
                expts.as_json(experiment_filename)
                if refls is not None:
                    reflections_filename = reflections_template(index=chunk_id)
                    print(
                        "Saving reflections for chunk %d to %s"
                        % (chunk_id, reflections_filename)
                    )
                    refls.as_file(reflections_filename)

            chunk_counter = 0
            chunk_expts = ExperimentList()
            if reflections:
                chunk_refls = flex.reflection_table()
            else:
                chunk_refls = None
            next_iset_id = 0
            imagesets_found = OrderedSet()
            for i, experiment in enumerate(experiments):
                chunk_expts.append(experiment)
                if reflections:
                    if reflections.experiment_identifiers().keys():
                        # first find which id value corresponds to experiment in question
                        identifier = experiment.identifier
                        id_ = None
                        for k in reflections.experiment_identifiers().keys():
                            if reflections.experiment_identifiers()[k] == identifier:
                                id_ = k
                                break
                        if id_ is None:
                            raise Sorry(
                                "Unable to find id matching experiment identifier in reflection table."
                            )
                        ref_sel = reflections.select(reflections["id"] == id_)
                        # now reset ids and reset/update identifiers map
                        for k in ref_sel.experiment_identifiers().keys():
                            del ref_sel.experiment_identifiers()[k]
                        new_id = len(chunk_expts) - 1
                        ref_sel["id"] = flex.int(len(ref_sel), new_id)
                        ref_sel.experiment_identifiers()[new_id] = identifier
                    else:
                        ref_sel = reflections.select(reflections["id"] == i)
                        ref_sel["id"] = flex.int(len(ref_sel), len(chunk_expts) - 1)
                    if experiment.imageset not in imagesets_found:
                        imagesets_found.add(experiment.imageset)
                        ref_sel["imageset_id"] = flex.int(ref_sel.size(), next_iset_id)
                        next_iset_id += 1
                    else:
                        iset_id = imagesets_found.index(experiment.imageset)
                        ref_sel["imageset_id"] = flex.int(ref_sel.size(), iset_id)
                    chunk_refls.extend(ref_sel)
                if params.output.chunk_sizes:
                    chunk_limit = params.output.chunk_sizes[chunk_counter]
                else:
                    chunk_limit = params.output.chunk_size
                if len(chunk_expts) == chunk_limit:
                    save_chunk(chunk_counter, chunk_expts, chunk_refls)
                    chunk_counter += 1
                    chunk_expts = ExperimentList()
                    if reflections:
                        chunk_refls = flex.reflection_table()
                    else:
                        chunk_refls = None
            if len(chunk_expts) > 0:
                save_chunk(chunk_counter, chunk_expts, chunk_refls)
        else:
            for i, experiment in enumerate(experiments):

                experiment_filename = experiments_template(index=i)
                print("Saving experiment %d to %s" % (i, experiment_filename))
                ExperimentList([experiment]).as_json(experiment_filename)

                if reflections is not None:
                    reflections_filename = reflections_template(index=i)
                    print(
                        "Saving reflections for experiment %d to %s"
                        % (i, reflections_filename)
                    )
                    ref_sel = reflections.select(reflections["id"] == i)
                    if ref_sel.experiment_identifiers().keys():
                        identifier = ref_sel.experiment_identifiers()[i]
                        for k in ref_sel.experiment_identifiers().keys():
                            del ref_sel.experiment_identifiers()[k]
                        ref_sel["id"] = flex.int(ref_sel.size(), 0)
                        ref_sel.experiment_identifiers()[0] = identifier
                    else:
                        ref_sel["id"] = flex.int(len(ref_sel), 0)
                    ref_sel["imageset_id"] = flex.int(len(ref_sel), 0)
                    ref_sel.as_file(reflections_filename)

        return


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



"""
Test prediction of reflections using the scan-varying reflection
predictor.
"""


from __future__ import annotations

import math

from cctbx.sgtbx import space_group, space_group_symbols
from dxtbx.model import ScanFactory
from dxtbx.model.experiment_list import Experiment, ExperimentList
from libtbx.phil import parse
from libtbx.test_utils import approx_equal
from scitbx.array_family import flex

from dials.algorithms.refinement.prediction.managed_predictors import (
    ScansExperimentsPredictor,
    ScansRayPredictor,
)
from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

from ..refinement import geometry_phil, setup_geometry


def setup_models(args):
    """setup the experimental models"""

    # Setup experimental models
    master_phil = parse(geometry_phil)

    models = setup_geometry.Extract(master_phil, cmdline_args=args)

    detector = models.detector
    goniometer = models.goniometer
    crystal = models.crystal
    beam = models.beam

    # Build a mock scan for a 180 degree sequence
    sf = ScanFactory()
    scan = sf.make_scan(
        image_range=(1, 180),
        exposure_times=0.1,
        oscillation=(0, 1.0),
        epochs=list(range(180)),
        deg=True,
    )
    sequence_range = scan.get_oscillation_range(deg=False)
    im_width = scan.get_oscillation(deg=False)[1]
    assert sequence_range == (0.0, math.pi)
    assert approx_equal(im_width, 1.0 * math.pi / 180.0)

    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            crystal=crystal,
            imageset=None,
        )
    )

    return experiments


def ref_gen_static(experiments):
    """Generate some reflections using the static predictor"""

    beam = experiments[0].beam
    crystal = experiments[0].crystal
    detector = experiments[0].detector
    scan = experiments[0].scan

    # All indices to the detector max resolution
    dmin = detector.get_max_resolution(beam.get_s0())
    index_generator = IndexGenerator(
        crystal.get_unit_cell(), space_group(space_group_symbols(1).hall()).type(), dmin
    )
    indices = index_generator.to_array()

    # Predict rays within the sequence range
    sequence_range = scan.get_oscillation_range(deg=False)
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    refs = ray_predictor(indices)

    # Take only those rays that intersect the detector
    intersects = ray_intersection(detector, refs)
    refs = refs.select(intersects)

    # Make a reflection predictor and re-predict for these reflections. The
    # result is the same, but we gain also the flags and xyzcal.px columns
    ref_predictor = ScansExperimentsPredictor(experiments)
    refs["id"] = flex.int(len(refs), 0)
    refs = ref_predictor(refs)

    return refs


def ref_gen_varying(experiments):
    """Generate some reflections using the scan varying predictor"""

    beam = experiments[0].beam
    crystal = experiments[0].crystal
    detector = experiments[0].detector
    scan = experiments[0].scan

    # We need a UB matrix at the beginning of every image, and at the end of the
    # last image. These are all the same - we want to compare the scan-varying
    # predictor with the scan-static one for a flat scan.
    ar_range = scan.get_array_range()
    UBlist = [crystal.get_A() for t in range(ar_range[0], ar_range[1] + 1)]
    dmin = detector.get_max_resolution(beam.get_s0())

    from dials.algorithms.spot_prediction import ScanVaryingReflectionPredictor

    sv_predictor = ScanVaryingReflectionPredictor(experiments[0], dmin=dmin)
    refs = sv_predictor.for_ub(flex.mat3_double(UBlist))

    return refs


def sort_refs(reflections):
    """Sort reflections by Miller index and entering flag"""

    refs_sorted = sorted(reflections.rows(), key=lambda x: x["entering"])
    refs_sorted = sorted(refs_sorted, key=lambda x: x["miller_index"][2])
    refs_sorted = sorted(refs_sorted, key=lambda x: x["miller_index"][1])
    refs_sorted = sorted(refs_sorted, key=lambda x: x["miller_index"][0])

    return refs_sorted


def test():
    experiments = setup_models([])
    refs1 = ref_gen_static(experiments)
    refs2 = ref_gen_varying(experiments)

    refs1_sorted = sort_refs(refs1)
    refs2_sorted = sort_refs(refs2)

    assert len(refs1_sorted) == len(refs2_sorted)

    for (r1, r2) in zip(refs1_sorted, refs2_sorted):
        assert r1["miller_index"] == r2["miller_index"]
        dz = r1["xyzcal.px"][2] - r2["xyzcal.px"][2]
        assert abs(dz) < 0.01



from __future__ import annotations

import json
import os

import procrunner
import pytest

import libtbx
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.serialize import load

import dials.util.masking
from dials.algorithms.shadowing.filter import filter_shadowed_reflections
from dials.array_family import flex


@pytest.mark.parametrize(
    "path,count_only_shadow,count_mask_shadow,count_mask_no_shadow",
    [
        (
            "shadow_test_data/DLS_I04_SmarGon/Th_3_O45_C45_P48_1_0500.cbf",
            pytest.approx(426758, abs=1e2),
            pytest.approx(917940, abs=1e2),
            pytest.approx(528032, abs=1e2),
        ),
        (
            "shadow_test_data/DLS_I03_SmarGon/protk_2_0600.cbf",
            pytest.approx(519100, abs=1e2),
            pytest.approx(1002068, abs=1e2),
            pytest.approx(527314, abs=1e2),
        ),
    ],
)
def test_dynamic_shadowing(
    path, count_only_shadow, count_mask_shadow, count_mask_no_shadow, dials_regression
):
    path = os.path.join(dials_regression, path)
    assert os.path.exists(path), path
    for shadowing in (libtbx.Auto, True, False):
        format_kwargs = {"dynamic_shadowing": shadowing}
        experiments = ExperimentListFactory.from_filenames(
            [path], format_kwargs=format_kwargs
        )
        imageset = experiments.imagesets()[0]
        detector = imageset.get_detector()
        scan = imageset.get_scan()
        masker = imageset.masker()
        if shadowing is not False:
            assert masker is not None
            mask = masker.get_mask(detector, scan.get_oscillation()[0])
            assert len(mask) == len(detector)
            # only shadowed pixels masked
            assert mask[0].count(False) == count_only_shadow, (
                mask[0].count(False),
                count_only_shadow,
            )
        mask = imageset.get_mask(0)

        # dead pixels, pixels in gaps, etc also masked
        if shadowing is libtbx.Auto or shadowing is True:
            assert mask[0].count(False) == count_mask_shadow, (
                mask[0].count(False),
                count_mask_shadow,
            )
        else:
            assert mask[0].count(False) == count_mask_no_shadow, (
                mask[0].count(False),
                count_mask_no_shadow,
            )


@pytest.mark.xfail(reason="Failing due to deprecation warning in output")
def test_shadow_plot(dials_data, tmp_path):
    result = procrunner.run(
        (
            "dials.import",
            dials_data("image_examples", pathlib=True) / "DLS_I03_smargon_0001.cbf.gz",
        ),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    result = procrunner.run(
        ("dials.shadow_plot", "imported.expt", "json=shadow.json"),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("scan_shadow_plot.png").is_file()
    assert tmp_path.joinpath("shadow.json").is_file()
    d = json.loads(tmp_path.joinpath("shadow.json").read_text())
    assert set(d) == {"fraction_shadowed", "scan_points"}
    assert d["fraction_shadowed"] == pytest.approx([0.003016, 0.003141], 2e-4)
    result = procrunner.run(
        ("dials.shadow_plot", "imported.expt", "mode=2d", "plot=shadow_2d.png"),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("shadow_2d.png").is_file()


def test_filter_shadowed_reflections(dials_regression):
    experiments_json = os.path.join(
        dials_regression, "shadow_test_data", "DLS_I04_SmarGon", "experiments.json"
    )
    predicted_pickle = os.path.join(
        dials_regression, "shadow_test_data", "DLS_I04_SmarGon", "predicted.pickle"
    )

    experiments = load.experiment_list(experiments_json, check_format=True)
    predicted = flex.reflection_table.from_file(predicted_pickle)

    for experiment_goniometer in (True, False):
        shadowed = filter_shadowed_reflections(
            experiments, predicted, experiment_goniometer=experiment_goniometer
        )
        assert shadowed.count(True) == 17
        assert shadowed.count(False) == 674


def test_lru_equality_cache_basic():
    callargs = []
    result = [None]

    def _callappend(*arg):
        callargs.append(arg)
        return result[-1]

    fun = dials.util.masking.lru_equality_cache(maxsize=20)(_callappend)
    hits, misses, maxsize, currsize = fun.cache_info()
    assert hits == 0
    assert misses == 0
    assert maxsize == 20
    assert currsize == 0

    for i in range(100):
        fun(i)
        fun(i)
    hits, misses, maxsize, currsize = fun.cache_info()
    assert hits == 100
    assert misses == 100
    assert maxsize == 20
    assert currsize == 20
    assert len(callargs) == 100
    assert len(set(callargs)) == 100
    for i in range(80, 100):
        fun(i)
    assert fun.cache_info() == (120, 100, 20, 20)
    assert len(callargs) == 100

    # Change the return value
    assert fun(1) is None
    result.append("test_result")
    assert fun(1) is None
    assert fun.__wrapped__(1) == "test_result"
    assert fun(2) == "test_result"


def test_lru_equality_cache_id():
    callargs = []

    def _callappend(*arg):
        callargs.append(arg)

    fun = dials.util.masking.lru_equality_cache(maxsize=1)(_callappend)

    class EqTester:
        def __init__(self, a):
            self.a = a

        def __eq__(self, other):
            return other.a == self.a

        def __hash__(self):
            return hash(id(self))

    a, b = EqTester(-424242), EqTester(-424242)
    assert a == b
    assert hash(a) != hash(b)

    fun(a)
    fun(b)
    assert fun.cache_info() == (1, 1, 1, 1)


def test_generate_mask(dials_data):
    imageset = load.imageset(
        dials_data("centroid_test_data", pathlib=True) / "sweep.json"
    )
    params = dials.util.masking.phil_scope.extract()
    params.border = 2
    params.d_min = 1.5
    params.d_max = 40
    params.untrusted[0].rectangle = (500, 600, 700, 800)
    params.untrusted[0].circle = (1000, 1200, 100)
    params.untrusted[0].polygon = (1500, 1500, 1600, 1600, 1700, 1500, 1500, 1500)
    params.untrusted[0].pixel = (1183, 1383)
    params.resolution_range = [(3.1, 3.2), (5.1, 5.2)]
    params.ice_rings.filter = True
    mask = dials.util.masking.generate_mask(imageset, params)
    assert len(mask) == len(imageset.get_detector())
    for m, im in zip(mask, imageset.get_raw_data(0)):
        assert m.all() == im.all()
    assert mask[0].count(False) == 4060817


@pytest.mark.parametrize(
    "disable_parallax_correction,expected", [(False, 1427394), (True, 1432002)]
)
def test_generate_mask_parallax_correction(
    disable_parallax_correction, expected, dials_data
):
    expts = ExperimentListFactory.from_filenames(
        sorted(dials_data("centroid_test_data", pathlib=True).glob("*.cbf"))
    )
    imageset = expts[0].imageset
    params = dials.util.masking.phil_scope.extract()
    params.disable_parallax_correction = disable_parallax_correction
    params.d_min = 1.2
    params.d_max = 40
    params.untrusted = []
    mask = dials.util.masking.generate_mask(imageset, params)
    assert len(mask) == len(imageset.get_detector())
    for m, im in zip(mask, imageset.get_raw_data(0)):
        assert m.all() == im.all()
    assert mask[0].count(False) == expected



from __future__ import annotations

import pytest

from cctbx import sgtbx
from dxtbx.model import Crystal
from scitbx import matrix
from scitbx.math import euler_angles as euler

from dials.algorithms.indexing import compare_orientation_matrices


def test_compare_orientation_matrices():
    # try and see if we can get back the original rotation matrix and euler angles
    real_space_a = matrix.col((10, 0, 0))
    real_space_b = matrix.col((0, 10, 10))
    real_space_c = matrix.col((0, 0, 10))
    euler_angles = (1.3, 5.6, 7.8)
    R = matrix.sqr(euler.xyz_matrix(euler_angles[0], euler_angles[1], euler_angles[2]))
    crystal_a = Crystal(
        real_space_a, real_space_b, real_space_c, space_group=sgtbx.space_group("P 1")
    )
    crystal_b = Crystal(
        R * real_space_a,
        R * real_space_b,
        R * real_space_c,
        space_group=sgtbx.space_group("P 1"),
    )
    assert (
        matrix.sqr(crystal_b.get_U()) * matrix.sqr(crystal_a.get_U()).transpose()
    ).elems == pytest.approx(R.elems)
    (
        best_R_ab,
        best_axis,
        best_angle,
        best_cb_op,
    ) = compare_orientation_matrices.difference_rotation_matrix_axis_angle(
        crystal_a, crystal_b
    )
    best_euler_angles = euler.xyz_angles(best_R_ab)
    assert best_euler_angles == pytest.approx(euler_angles)
    assert best_cb_op.is_identity_op()
    assert best_R_ab.elems == pytest.approx(R.elems)

    # now see if we can deconvolute the original euler angles after applying
    # a change of basis to one of the crystals
    crystal_a = Crystal(
        real_space_a, real_space_b, real_space_c, space_group=sgtbx.space_group("I 2 3")
    )
    cb_op = sgtbx.change_of_basis_op("z,x,y")
    crystal_b = Crystal(
        R * real_space_a,
        R * real_space_b,
        R * real_space_c,
        space_group=sgtbx.space_group("I 2 3"),
    ).change_basis(cb_op)
    (
        best_R_ab,
        best_axis,
        best_angle,
        best_cb_op,
    ) = compare_orientation_matrices.difference_rotation_matrix_axis_angle(
        crystal_a, crystal_b
    )
    best_euler_angles = euler.xyz_angles(best_R_ab)
    assert best_euler_angles == pytest.approx(euler_angles)
    assert best_cb_op.c() == cb_op.inverse().c()
    assert best_R_ab.elems == pytest.approx(R.elems)

    crystal_c = crystal_b.change_basis(sgtbx.change_of_basis_op("-y,-z,x"))
    assert crystal_c != crystal_b

    s = compare_orientation_matrices.rotation_matrix_differences(
        [crystal_a, crystal_b, crystal_c], comparison="pairwise"
    )
    s = "\n".join(s.splitlines()[:-1]).replace("-0.000", "0.000")
    print(s)
    assert (
        s
        == """\
Change of basis op: b,c,a
Rotation matrix to transform crystal 1 to crystal 2:
{{0.986, -0.135, 0.098},
 {0.138, 0.990, -0.023},
 {-0.094, 0.036, 0.995}}
Rotation of 9.738 degrees about axis (0.172, 0.565, 0.807)

Change of basis op: -a,-b,c
Rotation matrix to transform crystal 1 to crystal 3:
{{0.986, -0.135, 0.098},
 {0.138, 0.990, -0.023},
 {-0.094, 0.036, 0.995}}
Rotation of 9.738 degrees about axis (0.172, 0.565, 0.807)

Change of basis op: c,-a,-b
Rotation matrix to transform crystal 2 to crystal 3:
{{1.000, 0.000, 0.000},
 {0.000, 1.000, 0.000},
 {0.000, 0.000, 1.000}}"""
    )

    s = compare_orientation_matrices.rotation_matrix_differences(
        [crystal_a, crystal_b, crystal_c], comparison="sequential"
    )
    s = "\n".join(s.splitlines()[:-1]).replace("-0.000", "0.000")
    print(s)
    assert (
        s
        == """\
Change of basis op: b,c,a
Rotation matrix to transform crystal 1 to crystal 2:
{{0.986, -0.135, 0.098},
 {0.138, 0.990, -0.023},
 {-0.094, 0.036, 0.995}}
Rotation of 9.738 degrees about axis (0.172, 0.565, 0.807)

Change of basis op: c,-a,-b
Rotation matrix to transform crystal 2 to crystal 3:
{{1.000, 0.000, 0.000},
 {0.000, 1.000, 0.000},
 {0.000, 0.000, 1.000}}"""
    )

    s = compare_orientation_matrices.rotation_matrix_differences(
        (crystal_a, crystal_b), miller_indices=((1, 0, 0), (1, 1, 0))
    )
    assert (
        s
        == """\
Change of basis op: b,c,a
Rotation matrix to transform crystal 1 to crystal 2:
{{0.986, -0.135, 0.098},
 {0.138, 0.990, -0.023},
 {-0.094, 0.036, 0.995}}
Rotation of 9.738 degrees about axis (0.172, 0.565, 0.807)
(1,0,0): 15.26 deg
(1,1,0): 9.12 deg
"""
    )



"""
Correct integrated intensities to account for attenuation by a diamond anvil cell.

High pressure X-ray diffraction experiments often involve a diamond anvil pressure
cell, in which the sample is sandwiched between two anvils, effectively parallel flat
plates of diamond.  The passage of the incident and diffracted beam through the
anvils results in attenuation of both beams by the diamond by an amount that is
dependent on the path length of each beam through each anvil.

This utility calculates these path lengths and boosts the integrated reflection
intensities to remove the calculated effect of the diamond attenuation.

It is intended that this program be used to correct reflection intensities after
integration but before scaling.  Call it on the output of dials.integrate.

Examples::

  dials.anvil_correction integrated.expt integrated.refl

  dials.anvil_correction integrated.expt integrated.refl thickness=1.2 normal=1,0,0
"""


from __future__ import annotations

import logging
import sys
from typing import List, Sequence, SupportsFloat

import numpy as np
from scipy.spatial.transform import Rotation

import libtbx.phil
from cctbx.eltbx import attenuation_coefficient
from dxtbx.model import Experiment

import dials.util.log
from dials.array_family import flex
from dials.util.multi_dataset_handling import (
    parse_multiple_datasets,
    sort_tables_to_experiments_order,
)
from dials.util.options import ArgumentParser, flatten_experiments, flatten_reflections

Vector = Sequence[SupportsFloat]


logger = logging.getLogger("dials.command_line.anvil_correction")

phil_scope = libtbx.phil.parse(
    """
    anvil
        .caption = "Properties of the mounted diamond anvils"
    {
        density = 3510
            .type = float
            .help = "The density of the anvil material in kg per cubic metre.  " \
                    "The default is the typical density of synthetic diamond."

        thickness = 1.5925
            .type = float
            .help = "The thickness in mm of each anvil in the pressure cell.  " \
                    "The default is the thickness of the pressure cells in use on " \
                    "beamline I19 at Diamond Light Source."

        normal = 0, 1, 0
            .type = floats(size=3)
            .help = "A 3-vector orthogonal to the anvil surfaces in the laboratory " \
                    "frame when the goniometer is at zero datum, i.e. the axes are " \
                    "all at zero degrees.  The vector may be given un-normalised."
    }

    output {
        experiments = None
            .type = path
            .help = "The output experiment list file name. If None, don't output an " \
                    "experiment list file."
        reflections = corrected.refl
            .type = path
            .help = "The output reflection table file."
        log = dials.anvil_correction.log
            .type = path
    }
    """
)

# Get the tabulated NIST mass attenuation coefficient data for carbon.
carbon_attenuation_data = attenuation_coefficient.get_table("C")


def goniometer_rotation(
    experiment: Experiment, reflections: flex.reflection_table
) -> Rotation:
    """
    Calculate the goniometer rotation operator for each reflection.

    For each reflection, find the rotation operator that describes the position in
    the lab frame that the sample was in when the reflection was measured.

    Following the DXTBX model of a goniometer, whereby a scan is only possible
    around one physical axis at a time, the goniostat rotation operator can be
    represented as
      R = S ∘ R' ∘ F.
    Here:
        * S is the static 'setting rotation', the operator denoting the position of all
        parent axes of the scan axis, which hence defines the orientation of the scan
        axis;
        * R' is the the operator denoting the scan rotation.  It has a different
        value for each reflection, recording the scan position corresponding to each
        reflection centroid;
        * F is the static 'fixed rotation', denoting the orientation of all child axes
        of the scan axis.

    Args:
        experiment:  The DXTBX experiment object corresponding to the scan.
        reflections:  A table of reflections at which to calculate the rotations.

    Returns:
        An array of rotation operators, one per reflection in the reflection table.
    """
    # Get the axis of the scan rotation.
    rotation_axis = experiment.goniometer.get_rotation_axis_datum()
    # For each reflection, get the angle of the scan rotation.
    angles = reflections["xyzobs.mm.value"].parts()[2]
    # Construct a rotation vector (parallel with the rotation axis and with
    # magnitude equal to the rotation angle) for each reflection.
    # The shape of this array is (N, 3), where N is the number of reflections.
    rotvecs = np.outer(angles, rotation_axis)
    # Create a rotation operator for each scan rotation (i.e. one per reflection).
    # In the notation of dxtbx/model/goniometer.h, this is R.
    scan_rotation = Rotation.from_rotvec(rotvecs)

    # Get the setting rotation.
    # In the notation of dxtbx/model/goniometer.h, this is S.
    set_rotation = np.array(experiment.goniometer.get_setting_rotation()).reshape(3, 3)
    set_rotation = Rotation.from_matrix(set_rotation)

    # Create a rotation operator for those axes that are fixed throughout the scan.
    # In the notation of dxtbx/model/goniometer.h, this is F.
    fixed_rotation = np.array(experiment.goniometer.get_fixed_rotation()).reshape(3, 3)
    fixed_rotation = Rotation.from_matrix(fixed_rotation)

    # Calculate the rotation operator representing the goniometer orientation for each
    # reflection.  In the notation of dxtbx/model/goniometer.h this is S ∘ R ∘ F.
    return set_rotation * scan_rotation * fixed_rotation


def attenuation_correction(
    experiment: Experiment,
    reflections: flex.reflection_table,
    dac_norm: Vector,
    thickness: float,
    density: float,
) -> flex.double:
    """
    Calculate the correction factors for attenuation by a diamond anvil cell.

    Take an experiment object and reflection table containing integrated but unscaled
    diffraction data, estimated the factor by which each of the integrated
    reflection intensities must be boosted to correct for attenuation of the incident
    and diffracted beams in the diamond anvils.

    Args:
        experiment:  An experiment from integrated data.
        reflections:  A table of integrated reflections.
        dac_norm:  A 3-vector representing the normal to the anvil surfaces in the
                   laboratory frame when the goniometer is at zero datum, i.e. the axes
                   are all at 0°.  The vector is assumed to be normalised.
        thickness:  The thickness of each diamond anvil (assumed equal).
        density:  The density of the anvil material in g·cm⁻³
                  (units chosen for consistency with the NIST tables of X-ray mass
                  attenuation coefficients, https://dx.doi.org/10.18434/T4D01F).

    Returns:
        An array of correction factors for the integrated reflection intensities.
    """
    # Get the wavelength.
    wavelength = experiment.beam.get_wavelength()
    # Get the mass attenuation coefficient in cm²·g⁻¹ from the NIST tables.
    mass_atten_coeff = carbon_attenuation_data.mu_rho_at_angstrom(wavelength)
    # Get the linear attenuation coefficient in mm⁻¹.
    linear_atten_coeff = density * mass_atten_coeff / 10  # mm⁻¹

    # Find the orientation of the diamond anvil cell for each reflection.  Since the
    # cell is fixed to the goniometer, this is just S × R × F × ̠̂n.
    # This is an array of shape (N, 3), i.e. a 3-vector for each reflection.
    dac_orientation = goniometer_rotation(experiment, reflections).apply(dac_norm)

    # Get the normalised incident and diffracted beam vectors ̠̂s₀ and ̠̂s₁.
    # Naturally, there will be one vale of ̠̂s₁ for each reflection.
    s0_norm = experiment.beam.get_unit_s0()  # An array of shape (3).
    s1_norm = reflections["s1"] * wavelength  # Shape (N, 3).

    # Get the scalar product of the diamond anvil cell orientation with each of
    # ̠̂s₀ & ̠̂s₁ for each reflection.
    incident_cosine = np.dot(dac_orientation, s0_norm)  # An array of shape (N).
    # Dot product of each of an array of N 3-vectors with its counterpart from another
    # array of N 3-vectors.  Shape (N, 3) • (N, 3) → N.
    # Equivalent to np.sum(dac_orientation * s1_norm, axis=1), but quicker.
    diffracted_cosine = np.einsum("ij,ij->i", dac_orientation, s1_norm)  # Shape (N)

    # Get the path length through the anvil of the incident and reflected beams.
    l0 = thickness / np.abs(incident_cosine)  # mm.  An array of shape (N).
    l1 = thickness / np.abs(diffracted_cosine)  # mm.  An array of shape (N).

    # Derive the factor by which we estimate the intensities to have been attenuated
    # by the anvils.  This is an array of shape (N).
    l_tot = l0 + l1
    return flex.double(np.exp(linear_atten_coeff * l_tot))


def correct_intensities_for_dac_attenuation(
    experiment: Experiment,
    reflections: flex.reflection_table,
    dac_norm: Vector,
    thickness: float,
    density: float = 3.51,
) -> None:
    """
    Boost integrated intensities to account for attenuation by a diamond anvil cell.

    Take an experiment object and reflection table containing integrated but unscaled
    diffraction data, boost the integrated reflection intensities to correct for the
    estimated attenuation due to the passage of the incident and diffracted beams
    through the diamond anvils.

    Args:
        experiment:  An experiment from integrated data.
        reflections:  A table of reflections from integrated data.
        dac_norm:  A 3-vector representing the normal to the anvil surfaces in the
                   laboratory frame when the goniometer is at zero datum, i.e. the axes
                   are all at 0°.  The vector is assumed to be normalised.
        thickness:  The thickness of each diamond anvil (assumed equal).
        density:  The density of the anvil material in g·cm⁻³
                  (units chosen for consistency with the NIST tables of X-ray mass
                  attenuation coefficients, https://dx.doi.org/10.18434/T4D01F).
                  Defaults to a typical value for synthetic diamond.
    """
    # Select only those reflections whose intensities after integration ought to be
    # meaningful.
    sel = reflections.get_flags(reflections.flags.integrated, all=False)
    sel = sel.iselection()
    refls_sel = reflections.select(sel)

    correction = attenuation_correction(
        experiment, refls_sel, dac_norm, thickness, density
    )

    # We need only correct non-null values for each integration method.
    prf_subsel = refls_sel.get_flags(refls_sel.flags.integrated_prf)
    sum_subsel = refls_sel.get_flags(refls_sel.flags.integrated_sum)

    # Correct the measured intensities and variances for this attenuation.
    methods = {"prf": prf_subsel, "sum": sum_subsel}
    quantities = {"value": correction, "variance": flex.pow2(correction)}
    for method, subsel in methods.items():
        setting_subsel = sel.select(subsel)
        for quantity, factor in quantities.items():
            col = f"intensity.{method}.{quantity}"
            corrected = (refls_sel[col] * factor).select(subsel)
            try:
                reflections[col].set_selected(setting_subsel, corrected)
            except KeyError:
                pass


@dials.util.show_mail_handle_errors()
def run(args: List[str] = None, phil: libtbx.phil.scope = phil_scope) -> None:
    """
    Run dials.anvil_correction as from the command line.

    Take integrated experiment lists and reflection tables and correct the all the
    integrated intensities for the estimated attenuation by the diamond anvils.

    Args:
        args: The arguments supplied by the user (default: sys.argv[1:]).
        phil: The PHIL scope definition (default: phil_scope, the master PHIL scope
              for this program).
    """
    usage = "dials.anvil_correction [options] integrated.expt integrated.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=__doc__,
    )

    params, options = parser.parse_args(args=args, show_diff_phil=False)

    # Log the difference between the PHIL scope definition and the active PHIL scope,
    # which will include the parsed user inputs.
    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n%s", diff_phil)

    # Check that at least one reflection table and experiment list have been provided.
    input_errors = []
    if not params.input.experiments:
        input_errors.append(
            "Please provide at least one valid experiment list (.expt) file."
        )
    if not params.input.reflections:
        input_errors.append(
            "Please provide at least one valid reflection table (.refl) file."
        )
    if input_errors:
        sys.exit("\n".join([parser.format_help()] + input_errors))

    if not np.linalg.norm(params.anvil.normal):
        sys.exit("It seems you have provided a surface normal vector with zero length.")

    # Check that the anvil surface normal really is normalised.
    dac_norm = params.anvil.normal / np.linalg.norm(params.anvil.normal)

    # Configure the logging.
    dials.util.log.config(options.verbose, logfile=params.output.log)

    # These functions are commonly used to collate the input.
    experiments = flatten_experiments(params.input.experiments)
    reflections_list = flatten_reflections(params.input.reflections)
    # Work around parse_multiple_datasets dropping unindexed reflections.
    unindexed = flex.reflection_table()
    for r_table in reflections_list:
        unindexed.extend(r_table.select(r_table["id"] == -1))
    # Get a single reflection table per experiment object.
    reflections_list = parse_multiple_datasets(reflections_list)
    reflections_list = sort_tables_to_experiments_order(reflections_list, experiments)

    # Record the density of diamond in g·cm⁻³ (for consistency with NIST tables,
    # https://doi.org/10.18434/T4D01F).
    density = params.anvil.density / 1000  # g·cm⁻³

    # Correct for the attenuation of the incident and diffracted beams by the diamond
    # anvil cell.
    logger.info(
        "Correcting integrated reflection intensities for attenuation by the diamond "
        "anvil cell."
    )
    for experiment, reflections in zip(experiments, reflections_list):
        correct_intensities_for_dac_attenuation(
            experiment, reflections, dac_norm, params.anvil.thickness, density
        )
    logger.info("Done.")

    # Do optional experiment list file output here.
    if params.output.experiments:
        logger.info("Writing the experiment list to %s", params.output.experiments)
        experiments.as_file(params.output.experiments)
    logger.info("Writing the reflection table to %s", params.output.reflections)
    # Collate reflections into a single reflection table and save it to file.
    reflections = unindexed
    for r_table in reflections_list:
        reflections.extend(r_table)
    del reflections_list
    reflections.as_file(params.output.reflections)


# Keep this minimal.  Try to keep the command-line behaviour neatly encapsulated in run.
if __name__ == "__main__":
    run()



from __future__ import annotations

from functools import reduce

from scitbx import matrix
from scitbx.array_family import flex

from dials.algorithms.refinement.parameterisation.model_parameters import (
    ModelParameterisation,
    Parameter,
)
from dials.algorithms.refinement.refinement_helpers import (
    PanelGroupCompose,
    dR_from_axis_and_angle,
    get_panel_groups_at_depth,
    get_panel_ids_at_root,
)


class DetectorMixin:
    """Mix-in class defining some functionality unique to detector
    parameterisations that can be shared by static and scan-varying versions"""

    @staticmethod
    def _init_core(detector, parameter_type=Parameter):
        """Calculate initial state and list of parameters, using the parameter_type
        callback to select between versions of the Parameter class"""

        # get some vectors we need from the Panel
        panel = detector[0]
        so = matrix.col(panel.get_origin())
        d1 = matrix.col(panel.get_fast_axis())
        d2 = matrix.col(panel.get_slow_axis())
        dn = matrix.col(panel.get_normal())

        # We choose the dorg vector to terminate in the centre of the Panel, and
        # the offset between the end of the dorg vector and the Panel origin is
        # a coordinate matrix with elements in the basis d1, d2, dn
        panel_lim = panel.get_image_size_mm()
        offset = matrix.col((-1.0 * panel_lim[0] / 2.0, -1.0 * panel_lim[1] / 2.0, 0.0))

        dorg = so - offset[0] * d1 - offset[1] * d2

        # Set up the initial state. There are multiple items of interest, so
        # use a dictionary here (note for a single panel model we can do
        # without the first 3 of these, but will need them for multiple panels)
        istate = {"d1": d1, "d2": d2, "dn": dn, "offset": offset}

        # set up the parameters.
        # distance from lab origin to detector model plane along its normal, in
        # initial orientation
        distance = panel.get_directed_distance()
        dist = parameter_type(distance, dn, "length (mm)", "Dist")

        # shift in the detector model plane to locate dorg, in initial
        # orientation
        shift = dorg - dn * distance
        shift1 = parameter_type(shift.dot(d1), d1, "length (mm)", "Shift1")
        shift2 = parameter_type(shift.dot(d2), d2, "length (mm)", "Shift2")

        # rotations of the plane through its origin about:
        # 1) axis normal to initial orientation
        # 2) d1 axis of initial orientation
        # 3) d2 axis of initial orientation
        tau1 = parameter_type(0, dn, "angle (mrad)", "Tau1")
        tau2 = parameter_type(0, d1, "angle (mrad)", "Tau2")
        tau3 = parameter_type(0, d2, "angle (mrad)", "Tau3")

        # build the parameter list in a specific,  maintained order
        p_list = [dist, shift1, shift2, tau1, tau2, tau3]

        return {"istate": istate, "p_list": p_list}

    def _compose_core(self, dist, shift1, shift2, tau1, tau2, tau3):

        # extract items from the initial state
        id1 = self._initial_state["d1"]
        id2 = self._initial_state["d2"]
        ioffset = self._initial_state["offset"]

        # convert angles to radians
        tau1rad = tau1.value / 1000.0
        tau2rad = tau2.value / 1000.0
        tau3rad = tau3.value / 1000.0

        # compose rotation matrices and their first order derivatives
        Tau1 = (tau1.axis).axis_and_angle_as_r3_rotation_matrix(tau1rad, deg=False)
        dTau1_dtau1 = dR_from_axis_and_angle(tau1.axis, tau1rad, deg=False)

        Tau2 = (tau2.axis).axis_and_angle_as_r3_rotation_matrix(tau2rad, deg=False)
        dTau2_dtau2 = dR_from_axis_and_angle(tau2.axis, tau2rad, deg=False)

        Tau3 = (tau3.axis).axis_and_angle_as_r3_rotation_matrix(tau3rad, deg=False)
        dTau3_dtau3 = dR_from_axis_and_angle(tau3.axis, tau3rad, deg=False)

        Tau32 = Tau3 * Tau2
        Tau321 = Tau32 * Tau1

        # Compose new state
        # =================
        # First the frame positioned at a distance from the lab origin
        P0 = dist.value * dist.axis  # distance along initial detector normal
        Px = P0 + id1  # point at the end of d1 in lab frame
        Py = P0 + id2  # point at the end of d2 in lab frame

        # detector shift vector
        dsv = P0 + shift1.value * shift1.axis + shift2.value * shift2.axis

        # compose dorg point
        dorg = Tau321 * dsv - Tau32 * P0 + P0

        # compose d1, d2 and dn and ensure frame remains orthonormal.
        d1 = (Tau321 * (Px - P0)).normalize()
        d2 = (Tau321 * (Py - P0)).normalize()
        dn = d1.cross(d2).normalize()
        # NB dn not actually used in this simple model; calculation left
        # here as a reminder for future extension
        d2 = dn.cross(d1)

        # compose new sensor origin
        o = dorg + ioffset[0] * d1 + ioffset[1] * d2

        # keep the new state for return
        new_state = {"d1": d1, "d2": d2, "origin": o}

        # calculate derivatives of the state wrt parameters
        # =================================================
        # Start with the dorg vector, where
        # dorg = Tau321 * dsv - Tau32 * P0 + P0

        # derivative wrt dist
        dP0_ddist = dist.axis
        ddsv_ddist = dP0_ddist
        ddorg_ddist = Tau321 * ddsv_ddist - Tau32 * dP0_ddist + dP0_ddist

        # derivative wrt shift1
        ddsv_dshift1 = shift1.axis
        ddorg_dshift1 = Tau321 * ddsv_dshift1

        # derivative wrt shift2
        ddsv_dshift2 = shift2.axis
        ddorg_dshift2 = Tau321 * ddsv_dshift2

        # derivative wrt tau1
        dTau321_dtau1 = Tau32 * dTau1_dtau1
        ddorg_dtau1 = dTau321_dtau1 * dsv

        # derivative wrt tau2
        dTau32_dtau2 = Tau3 * dTau2_dtau2
        dTau321_dtau2 = dTau32_dtau2 * Tau1
        ddorg_dtau2 = dTau321_dtau2 * dsv - dTau32_dtau2 * P0

        # derivative wrt tau3
        dTau32_dtau3 = dTau3_dtau3 * Tau2
        dTau321_dtau3 = dTau32_dtau3 * Tau1
        ddorg_dtau3 = dTau321_dtau3 * dsv - dTau32_dtau3 * P0

        # Now derivatives of the direction d1, where
        # d1 = (Tau321 * (Px - P0)).normalize()
        # For calc of derivatives ignore the normalize(), which should
        # be unnecessary anyway as Px - P0 is a unit vector and Tau321 a
        # pure rotation.

        # derivative wrt dist
        # dPx_ddist = dist.axis; dP0_ddist = dist.axis, so these cancel
        dd1_ddist = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift1
        dd1_dshift1 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift2
        dd1_dshift2 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt tau1
        dd1_dtau1 = dTau321_dtau1 * (Px - P0)

        # derivative wrt tau2
        dd1_dtau2 = dTau321_dtau2 * (Px - P0)

        # derivative wrt tau3
        dd1_dtau3 = dTau321_dtau3 * (Px - P0)

        # Derivatives of the direction d2, where
        # d2 = (Tau321 * (Py - P0)).normalize()

        # derivative wrt dist
        dd2_ddist = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift1
        dd2_dshift1 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift2
        dd2_dshift2 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt tau1
        dd2_dtau1 = dTau321_dtau1 * (Py - P0)

        # derivative wrt tau2
        dd2_dtau2 = dTau321_dtau2 * (Py - P0)

        # derivative wrt tau3
        dd2_dtau3 = dTau321_dtau3 * (Py - P0)

        # Derivatives of the direction dn, where dn = d1.cross(d2).normalize()
        # These derivatives are not used, but are left as comments for understanding

        # derivative wrt dist
        # ddn_ddist = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift1
        # ddn_dshift1 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift2
        # ddn_dshift2 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt tau1. Product rule for cross product applies
        # ddn_dtau1 = dd1_dtau1.cross(d2) + d1.cross(dd2_dtau1)

        # derivative wrt tau2
        # ddn_dtau2 = dd1_dtau2.cross(d2) + d1.cross(dd2_dtau2)

        # derivative wrt tau3
        # ddn_dtau3 = dd1_dtau3.cross(d2) + d1.cross(dd2_dtau3)

        # calculate derivatives of the attached sensor matrix
        # ===================================================
        # sensor origin:
        # o = dorg + ioffset[0] * d1 + ioffset[1] * d2

        # derivative wrt dist
        do_ddist = ddorg_ddist + ioffset[0] * dd1_ddist + ioffset[1] * dd2_ddist

        # derivative wrt shift1
        do_dshift1 = ddorg_dshift1 + ioffset[0] * dd1_dshift1 + ioffset[1] * dd2_dshift1

        # derivative wrt shift2
        do_dshift2 = ddorg_dshift2 + ioffset[0] * dd1_dshift2 + ioffset[1] * dd2_dshift2

        # derivative wrt tau1
        do_dtau1 = ddorg_dtau1 + ioffset[0] * dd1_dtau1 + ioffset[1] * dd2_dtau1

        # derivative wrt tau2
        do_dtau2 = ddorg_dtau2 + ioffset[0] * dd1_dtau2 + ioffset[1] * dd2_dtau2

        # derivative wrt tau3
        do_dtau3 = ddorg_dtau3 + ioffset[0] * dd1_dtau3 + ioffset[1] * dd2_dtau3

        # combine these vectors together into derivatives of the sensor
        # matrix d, converting angles back to mrad
        dd_dval = []

        # derivative wrt dist
        dd_dval.append(
            matrix.sqr(dd1_ddist.elems + dd2_ddist.elems + do_ddist.elems).transpose()
        )

        # derivative wrt shift1
        dd_dval.append(
            matrix.sqr(
                dd1_dshift1.elems + dd2_dshift1.elems + do_dshift1.elems
            ).transpose()
        )

        # derivative wrt shift2
        dd_dval.append(
            matrix.sqr(
                dd1_dshift2.elems + dd2_dshift2.elems + do_dshift2.elems
            ).transpose()
        )

        # derivative wrt tau1
        dd_dval.append(
            matrix.sqr(dd1_dtau1.elems + dd2_dtau1.elems + do_dtau1.elems).transpose()
            / 1000.0
        )

        # derivative wrt tau2
        dd_dval.append(
            matrix.sqr(dd1_dtau2.elems + dd2_dtau2.elems + do_dtau2.elems).transpose()
            / 1000.0
        )

        # derivative wrt tau3
        dd_dval.append(
            matrix.sqr(dd1_dtau3.elems + dd2_dtau3.elems + do_dtau3.elems).transpose()
            / 1000.0
        )

        return new_state, dd_dval


class DetectorParameterisationSinglePanel(ModelParameterisation, DetectorMixin):
    """A parameterisation for a Detector model with a single abstract panel plane.

    A single panel has 6 parameters: three translations (detector distance and
    two in-plane shifts) and three rotations (around the initial normal, fast and
    slow axes) with angles expressed in mrad.
    """

    def __init__(self, detector, experiment_ids=None):
        """Initialise the DetectorParameterisationSinglePanel object

        Args:
            detector: A dxtbx Detector object to be parameterised.
            experiment_ids (list): The experiment IDs affected by this
                parameterisation. Defaults to None, which is replaced by [0].
        """
        # The state of a single Panel is its detector matrix d = (d1|d2|d0).
        # However, for the purposes of parameterisation we choose a different
        # vector than d0 to locate the Panel. That's because we want to perform
        # rotations around a point on the detector surface, and d0 points to the
        # corner of the Panel. To avoid excess correlations between 'tilt' and
        # 'twist' angles with the detector distance, we prefer to perform
        # rotations around a point located at the centre of the Panel. This is
        # usually close to point of intersection with the plane normal drawn
        # from the origin of the laboratory frame.
        #
        # Therefore we define:
        #
        # * a vector 'dorg' locating the centre of the single Panel
        # * a pair of orthogonal unit directions 'd1' and 'd2' forming a plane
        #   with its origin at the end of the vector dorg
        # * a third unit direction 'dn', orthogonal to both 'd1' & 'd2'.
        # * offsets to locate the origin d0 of the Panel frame from the
        #   tip of the dorg vector, in terms of the coordinate system
        #   formed by d1, d2 and dn.
        #
        # Held separately in attribute 'models' are:
        # * references to the detector objects contained in this model
        #
        # For this simplified class there is only a single Panel frame
        # and the vector dn is not actually required, because the plane formed
        # by d1 and d2 is coplanar with the sensor plane. Therefore the
        # offset is fully in terms of d1 and d2

        # set up the initial state of the detector parameterisation from the
        # orientation of the single Panel it contains, in terms of the vectors
        # dorg, d1 and d2.

        if experiment_ids is None:
            experiment_ids = [0]

        dat = self._init_core(detector)

        # set up the base class
        ModelParameterisation.__init__(
            self, detector, dat["istate"], dat["p_list"], experiment_ids=experiment_ids
        )

        # call compose to calculate all the derivatives
        self.compose()

    def compose(self):

        # extract parameters from the internal list
        dist, shift1, shift2, tau1, tau2, tau3 = self._param

        new_state, self._dstate_dp = self._compose_core(
            dist, shift1, shift2, tau1, tau2, tau3
        )

        # now update the panel with its new position and orientation.
        # The detector is self._model, the panel is the first in the
        # detector
        (self._model)[0].set_frame(
            new_state["d1"], new_state["d2"], new_state["origin"]
        )

    def get_state(self):

        # only a single panel exists, so no multi_state_elt argument is allowed
        panel = (self._model)[0]
        return matrix.sqr(panel.get_d_matrix())


class DetectorParameterisationMultiPanel(ModelParameterisation):
    """A parameterisation for a Detector model with multiple panels.

    The whole detector is treated as a single rigid block with 6 degrees of
    freedom (see DetectorParameterisationSinglePanel). The axes of the
    translations and rotations are taken from the 'middle panel' of the detector
    (the one closest to the direct beam).
    """

    def __init__(self, detector, beam, experiment_ids=None):
        """Initialise the DetectorParameterisationMultiPanel object

        Args:
            detector: A dxtbx Detector object to be parameterised.
            beam: An dxtbx beam object used to calculate the closest panel.
            experiment_ids (list): The experiment IDs affected by this
                parameterisation. Defaults to None, which is replaced by [0].
        """
        # The state of each Panel in the detector model is its matrix
        # d = (d1|d2|d0). We need to define a new coordinate system rigidly
        # attached to the detector model in which to express the
        # parameterisation and compose each of the Panel states.
        #
        # We define:
        #
        # * a vector 'dorg' locating a point in laboratory space that moves with
        #   the rigid body of the detector and thus is fixed wrt each of the
        #   Panels.
        # * A pair of orthogonal unit directions 'd1' and 'd2' forming a plane
        #   with its origin at the end of the vector dorg.
        # * a third unit direction 'dn', orthogonal to both 'd1' & 'd2'.
        # * offsets to locate the origin of each panel frame from the
        #   tip of the dorg vector, in terms of the coordinate system
        #   formed by d1, d2 and dn.
        #
        # Held separately in attribute 'models' are:
        # * references to detector objects contained in this model

        # set up the initial state of the detector model from the
        # orientation of whichever Panel has its centre most closely
        # located to the direct beam intersection. Call this 'mid_panel'

        if experiment_ids is None:
            experiment_ids = [0]
        beam_centres = [
            matrix.col(p.get_beam_centre(beam.get_unit_s0())) for p in detector
        ]
        panel_centres = [0.5 * matrix.col(p.get_image_size_mm()) for p in detector]
        beam_to_centres = [
            (a - b).length() for a, b in zip(beam_centres, panel_centres)
        ]
        mid_panel_id = beam_to_centres.index(min(beam_to_centres))
        mid_panel = detector[mid_panel_id]

        # get some vectors we need from the mid_panel
        so = matrix.col(mid_panel.get_origin())
        d1 = matrix.col(mid_panel.get_fast_axis())
        d2 = matrix.col(mid_panel.get_slow_axis())
        dn = matrix.col(mid_panel.get_normal())

        # we choose the dorg vector to terminate in the centre of the mid_panel,
        # and the offset between the end of the dorg vector and each Panel
        # origin is a coordinate matrix with elements in the basis d1, d2, dn.
        # We need also each Panel's plane directions dir1 and dir2 in terms of
        # d1, d2 and dn.
        mid_panel_centre = panel_centres[mid_panel_id]
        dorg = so + mid_panel_centre[0] * d1 + mid_panel_centre[1] * d2

        offsets, dir1s, dir2s = [], [], []
        for p in detector:
            offset = matrix.col(p.get_origin()) - dorg
            offsets.append(matrix.col((offset.dot(d1), offset.dot(d2), offset.dot(dn))))
            dir1 = matrix.col(p.get_fast_axis())
            dir1_new_basis = matrix.col((dir1.dot(d1), dir1.dot(d2), dir1.dot(dn)))
            dir1s.append(dir1_new_basis)
            dir2 = matrix.col(p.get_slow_axis())
            dir2_new_basis = matrix.col((dir2.dot(d1), dir2.dot(d2), dir2.dot(dn)))
            dir2s.append(dir2_new_basis)

        # The offsets and directions in the d1, d2, dn basis are fixed
        # quantities, not dependent on parameter values.
        self._offsets = offsets
        self._dir1s = dir1s
        self._dir2s = dir2s

        # Set up the initial state. This is the basis d1, d2, dn.
        istate = {"d1": d1, "d2": d2, "dn": dn}

        # set up the parameters.
        # distance from lab origin to mid_panel plane along its normal,
        # in initial orientation
        distance = mid_panel.get_directed_distance()
        dist = Parameter(distance, dn, "length (mm)", "Dist")

        # shift in the detector model plane to locate dorg, in initial
        # orientation
        shift = dorg - dn * distance
        shift1 = Parameter(shift.dot(d1), d1, "length (mm)", "Shift1")
        shift2 = Parameter(shift.dot(d2), d2, "length (mm)", "Shift2")

        # rotations of the plane through its origin about:
        # 1) axis normal to initial orientation
        # 2) d1 axis of initial orientation
        # 3) d2 axis of initial orientation
        tau1 = Parameter(0, dn, "angle (mrad)", "Tau1")
        tau2 = Parameter(0, d1, "angle (mrad)", "Tau2")
        tau3 = Parameter(0, d2, "angle (mrad)", "Tau3")

        # build the parameter list in a specific,  maintained order
        p_list = [dist, shift1, shift2, tau1, tau2, tau3]

        # set up the base class
        ModelParameterisation.__init__(
            self,
            detector,
            istate,
            p_list,
            experiment_ids=experiment_ids,
            is_multi_state=True,
        )

        # call compose to calculate all the derivatives
        self.compose()

    def compose(self):

        # extract parameters from the internal list
        dist, shift1, shift2, tau1, tau2, tau3 = self._param

        # convert angles to radians
        tau1rad = tau1.value / 1000.0
        tau2rad = tau2.value / 1000.0
        tau3rad = tau3.value / 1000.0

        # compose rotation matrices and their first order derivatives
        Tau1 = (tau1.axis).axis_and_angle_as_r3_rotation_matrix(tau1rad, deg=False)
        dTau1_dtau1 = dR_from_axis_and_angle(tau1.axis, tau1rad, deg=False)

        Tau2 = (tau2.axis).axis_and_angle_as_r3_rotation_matrix(tau2rad, deg=False)
        dTau2_dtau2 = dR_from_axis_and_angle(tau2.axis, tau2rad, deg=False)

        Tau3 = (tau3.axis).axis_and_angle_as_r3_rotation_matrix(tau3rad, deg=False)
        dTau3_dtau3 = dR_from_axis_and_angle(tau3.axis, tau3rad, deg=False)

        # Compose the new state
        from scitbx.array_family import flex

        from dials_refinement_helpers_ext import multi_panel_compose

        ret = multi_panel_compose(
            flex.vec3_double([self._initial_state[tag] for tag in ("d1", "d2", "dn")]),
            flex.double([p.value for p in self._param]),
            flex.vec3_double([p.axis for p in self._param]),
            self._model,
            flex.vec3_double(self._offsets),
            flex.vec3_double(self._dir1s),
            flex.vec3_double(self._dir2s),
            Tau1,
            dTau1_dtau1,
            Tau2,
            dTau2_dtau2,
            Tau3,
            dTau3_dtau3,
        )

        # Store the results.  The results come back as a single array, convert it to a 2D array
        self._multi_state_derivatives = [
            [
                matrix.sqr(ret[j * len(self._offsets) + i])
                for j in range(len(self._param))
            ]
            for i in range(len(self._offsets))
        ]

    def get_state(self, multi_state_elt=0):

        # There is only one detector, but the req. panel must be specified
        panel = (self._model)[multi_state_elt]
        return matrix.sqr(panel.get_d_matrix())


class PyDetectorParameterisationMultiPanel(DetectorParameterisationMultiPanel):
    """A python only version of DetectorParameterisationMultiPanel

    This version uses an older python-only compose method for comparison. See
    the base class for more details"""

    def compose(self):

        # extract items from the initial state
        id1 = self._initial_state["d1"]
        id2 = self._initial_state["d2"]

        # extract parameters from the internal list
        dist, shift1, shift2, tau1, tau2, tau3 = self._param

        # Extract the detector model
        detector = self._model

        # convert angles to radians
        tau1rad = tau1.value / 1000.0
        tau2rad = tau2.value / 1000.0
        tau3rad = tau3.value / 1000.0

        # compose rotation matrices and their first order derivatives
        Tau1 = (tau1.axis).axis_and_angle_as_r3_rotation_matrix(tau1rad, deg=False)
        dTau1_dtau1 = dR_from_axis_and_angle(tau1.axis, tau1rad, deg=False)

        Tau2 = (tau2.axis).axis_and_angle_as_r3_rotation_matrix(tau2rad, deg=False)
        dTau2_dtau2 = dR_from_axis_and_angle(tau2.axis, tau2rad, deg=False)

        Tau3 = (tau3.axis).axis_and_angle_as_r3_rotation_matrix(tau3rad, deg=False)
        dTau3_dtau3 = dR_from_axis_and_angle(tau3.axis, tau3rad, deg=False)

        Tau32 = Tau3 * Tau2
        Tau321 = Tau32 * Tau1

        # Compose new state
        # =================
        # First the frame positioned at a distance from the lab origin
        P0 = dist.value * dist.axis  # distance along initial detector normal
        Px = P0 + id1  # point at the end of d1 in lab frame
        Py = P0 + id2  # point at the end of d2 in lab frame

        # detector shift vector
        dsv = P0 + shift1.value * shift1.axis + shift2.value * shift2.axis

        # compose dorg point
        dorg = Tau321 * dsv - Tau32 * P0 + P0

        # compose new d1, d2 and dn and ensure frame remains orthonormal.
        d1 = (Tau321 * (Px - P0)).normalize()
        d2 = (Tau321 * (Py - P0)).normalize()
        dn = d1.cross(d2).normalize()
        d2 = dn.cross(d1)

        # compose new Panel origins
        origins = [
            dorg + offset[0] * d1 + offset[1] * d2 + offset[2] * dn
            for offset in self._offsets
        ]

        # compose new Panel directions
        dir1s = [vec[0] * d1 + vec[1] * d2 + vec[2] * dn for vec in self._dir1s]
        dir2s = [vec[0] * d1 + vec[1] * d2 + vec[2] * dn for vec in self._dir2s]

        # now update the panels with their new position and orientation.
        for p, dir1, dir2, org in zip(detector, dir1s, dir2s, origins):

            p.set_frame(dir1, dir2, org)

        # calculate derivatives of the state wrt parameters
        # =================================================
        # Start with the dorg vector, where
        # dorg = Tau321 * dsv - Tau32 * P0 + P0

        # derivative wrt dist
        dP0_ddist = dist.axis
        ddsv_ddist = dP0_ddist
        ddorg_ddist = Tau321 * ddsv_ddist - Tau32 * dP0_ddist + dP0_ddist

        # derivative wrt shift1
        ddsv_dshift1 = shift1.axis
        ddorg_dshift1 = Tau321 * ddsv_dshift1

        # derivative wrt shift2
        ddsv_dshift2 = shift2.axis
        ddorg_dshift2 = Tau321 * ddsv_dshift2

        # derivative wrt tau1
        dTau321_dtau1 = Tau32 * dTau1_dtau1
        ddorg_dtau1 = dTau321_dtau1 * dsv

        # derivative wrt tau2
        dTau32_dtau2 = Tau3 * dTau2_dtau2
        dTau321_dtau2 = dTau32_dtau2 * Tau1
        ddorg_dtau2 = dTau321_dtau2 * dsv - dTau32_dtau2 * P0

        # derivative wrt tau3
        dTau32_dtau3 = dTau3_dtau3 * Tau2
        dTau321_dtau3 = dTau32_dtau3 * Tau1
        ddorg_dtau3 = dTau321_dtau3 * dsv - dTau32_dtau3 * P0

        # Now derivatives of the direction d1, where
        # d1 = (Tau321 * (Px - P0)).normalize()
        # For calc of derivatives ignore the normalize(), which should
        # be unnecessary anyway as Px - P0 is a unit vector and Tau321 a
        # pure rotation.

        # derivative wrt dist
        # dPx_ddist = dist.axis; dP0_ddist = dist.axis, so these cancel
        dd1_ddist = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift1
        dd1_dshift1 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift2
        dd1_dshift2 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt tau1
        dd1_dtau1 = dTau321_dtau1 * (Px - P0)

        # derivative wrt tau2
        dd1_dtau2 = dTau321_dtau2 * (Px - P0)

        # derivative wrt tau3
        dd1_dtau3 = dTau321_dtau3 * (Px - P0)

        # Derivatives of the direction d2, where
        # d2 = (Tau321 * (Py - P0)).normalize()

        # derivative wrt dist
        dd2_ddist = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift1
        dd2_dshift1 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift2
        dd2_dshift2 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt tau1
        dd2_dtau1 = dTau321_dtau1 * (Py - P0)

        # derivative wrt tau2
        dd2_dtau2 = dTau321_dtau2 * (Py - P0)

        # derivative wrt tau3
        dd2_dtau3 = dTau321_dtau3 * (Py - P0)

        # Derivatives of the direction dn, where
        # dn = d1.cross(d2).normalize()

        # derivative wrt dist
        ddn_ddist = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift1
        ddn_dshift1 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt shift2
        ddn_dshift2 = matrix.col((0.0, 0.0, 0.0))

        # derivative wrt tau1. Product rule for cross product applies
        ddn_dtau1 = dd1_dtau1.cross(d2) + d1.cross(dd2_dtau1)

        # derivative wrt tau2
        ddn_dtau2 = dd1_dtau2.cross(d2) + d1.cross(dd2_dtau2)

        # derivative wrt tau3
        ddn_dtau3 = dd1_dtau3.cross(d2) + d1.cross(dd2_dtau3)

        # reset stored derivatives
        for i in range(len(detector)):
            self._multi_state_derivatives[i] = [None] * len(self._dstate_dp)

        # calculate derivatives of the attached Panel matrices
        # ====================================================
        for panel_id, (offset, dir1_new_basis, dir2_new_basis) in enumerate(
            zip(self._offsets, self._dir1s, self._dir2s)
        ):

            # Panel origin:
            # o = dorg + offset[0] * d1 + offset[1] * d2 + offset[2] * dn

            # derivative wrt dist. NB only ddorg_ddist is not null! The other
            # elements are left here to aid understanding, but should be removed
            # when this class is ported to C++ for speed.
            do_ddist = (
                ddorg_ddist
                + offset[0] * dd1_ddist
                + offset[1] * dd2_ddist
                + offset[2] * ddn_ddist
            )

            # derivative wrt shift1. NB only ddorg_dshift1 is non-null.
            do_dshift1 = (
                ddorg_dshift1
                + offset[0] * dd1_dshift1
                + offset[1] * dd2_dshift1
                + offset[2] * ddn_dshift1
            )

            # derivative wrt shift2. NB only ddorg_dshift2 is non-null.
            do_dshift2 = (
                ddorg_dshift2
                + offset[0] * dd1_dshift2
                + offset[1] * dd2_dshift2
                + offset[2] * ddn_dshift2
            )

            # derivative wrt tau1
            do_dtau1 = (
                ddorg_dtau1
                + offset[0] * dd1_dtau1
                + offset[1] * dd2_dtau1
                + offset[2] * ddn_dtau1
            )

            # derivative wrt tau2
            do_dtau2 = (
                ddorg_dtau2
                + offset[0] * dd1_dtau2
                + offset[1] * dd2_dtau2
                + offset[2] * ddn_dtau2
            )

            # derivative wrt tau3
            do_dtau3 = (
                ddorg_dtau3
                + offset[0] * dd1_dtau3
                + offset[1] * dd2_dtau3
                + offset[2] * ddn_dtau3
            )

            # Panel dir1:
            # dir1 = dir1_new_basis[0] * d1 + dir1_new_basis[1] * d2 +
            #        dir1_new_basis[2] * dn

            # derivative wrt dist. NB These are all null.
            ddir1_ddist = (
                dir1_new_basis[0] * dd1_ddist
                + dir1_new_basis[1] * dd2_ddist
                + dir1_new_basis[2] * ddn_ddist
            )

            # derivative wrt shift1. NB These are all null.
            ddir1_dshift1 = (
                dir1_new_basis[0] * dd1_dshift1
                + dir1_new_basis[1] * dd2_dshift1
                + dir1_new_basis[2] * ddn_dshift1
            )

            # derivative wrt shift2. NB These are all null.
            ddir1_dshift2 = (
                dir1_new_basis[0] * dd1_dshift2
                + dir1_new_basis[1] * dd2_dshift2
                + dir1_new_basis[2] * ddn_dshift2
            )

            # derivative wrt tau1
            ddir1_dtau1 = (
                dir1_new_basis[0] * dd1_dtau1
                + dir1_new_basis[1] * dd2_dtau1
                + dir1_new_basis[2] * ddn_dtau1
            )

            # derivative wrt tau2
            ddir1_dtau2 = (
                dir1_new_basis[0] * dd1_dtau2
                + dir1_new_basis[1] * dd2_dtau2
                + dir1_new_basis[2] * ddn_dtau2
            )

            # derivative wrt tau3
            ddir1_dtau3 = (
                dir1_new_basis[0] * dd1_dtau3
                + dir1_new_basis[1] * dd2_dtau3
                + dir1_new_basis[2] * ddn_dtau3
            )

            # Panel dir2:
            # dir2 = dir2_new_basis[0] * d1 + dir2_new_basis[1] * d2 +
            #        dir2_new_basis[2] * dn

            # derivative wrt dist. NB These are all null.
            ddir2_ddist = (
                dir2_new_basis[0] * dd1_ddist
                + dir2_new_basis[1] * dd2_ddist
                + dir2_new_basis[2] * ddn_ddist
            )

            # derivative wrt shift1. NB These are all null.
            ddir2_dshift1 = (
                dir2_new_basis[0] * dd1_dshift1
                + dir2_new_basis[1] * dd2_dshift1
                + dir2_new_basis[2] * ddn_dshift1
            )

            # derivative wrt shift2. NB These are all null.
            ddir2_dshift2 = (
                dir2_new_basis[0] * dd1_dshift2
                + dir2_new_basis[1] * dd2_dshift2
                + dir2_new_basis[2] * ddn_dshift2
            )

            # derivative wrt tau1
            ddir2_dtau1 = (
                dir2_new_basis[0] * dd1_dtau1
                + dir2_new_basis[1] * dd2_dtau1
                + dir2_new_basis[2] * ddn_dtau1
            )

            # derivative wrt tau2
            ddir2_dtau2 = (
                dir2_new_basis[0] * dd1_dtau2
                + dir2_new_basis[1] * dd2_dtau2
                + dir2_new_basis[2] * ddn_dtau2
            )

            # derivative wrt tau3
            ddir2_dtau3 = (
                dir2_new_basis[0] * dd1_dtau3
                + dir2_new_basis[1] * dd2_dtau3
                + dir2_new_basis[2] * ddn_dtau3
            )

            # combine these vectors together into derivatives of the panel
            # matrix d and store them, converting angles back to mrad
            self._multi_state_derivatives[panel_id] = [
                matrix.sqr(
                    ddir1_ddist.elems + ddir2_ddist.elems + do_ddist.elems
                ).transpose(),
                matrix.sqr(
                    ddir1_dshift1.elems + ddir2_dshift1.elems + do_dshift1.elems
                ).transpose(),
                matrix.sqr(
                    ddir1_dshift2.elems + ddir2_dshift2.elems + do_dshift2.elems
                ).transpose(),
                matrix.sqr(
                    ddir1_dtau1.elems + ddir2_dtau1.elems + do_dtau1.elems
                ).transpose()
                / 1000.0,
                matrix.sqr(
                    ddir1_dtau2.elems + ddir2_dtau2.elems + do_dtau2.elems
                ).transpose()
                / 1000.0,
                matrix.sqr(
                    ddir1_dtau3.elems + ddir2_dtau3.elems + do_dtau3.elems
                ).transpose()
                / 1000.0,
            ]


class DetectorParameterisationHierarchical(DetectorParameterisationMultiPanel):
    """A parameterisation for a hierarchical Detector model with multiple panels.

    The detector hierarchy is used to determine panel groups, each of which will
    be treated as a single rigid block with 6 degrees of freedom (see
    DetectorParameterisationSinglePanel).
    """

    def __init__(self, detector, experiment_ids=None, level=0):
        """Initialise the DetectorParameterisationHierarchical object

        Args:
            detector: A dxtbx Detector object to be parameterised.
            experiment_ids (list): The experiment IDs affected by this
                parameterisation. Defaults to None, which is replaced by [0].
            level (int): Select level of the detector hierarchy to determine panel
                groupings that are treated as separate rigid blocks.
        """

        if experiment_ids is None:
            experiment_ids = [0]

        try:
            h = detector.hierarchy()
        except AttributeError:
            print("This detector does not have a hierarchy")
            raise

        # list the panel groups at the chosen level
        try:
            self._groups = get_panel_groups_at_depth(h, level)
        except AttributeError:
            print(f"Cannot access the hierarchy at the depth level={level}")
            raise

        # collect the panel ids for each Panel within the groups
        panels = list(detector)
        self._panel_ids_by_group = [
            get_panel_ids_at_root(panels, g) for g in self._groups
        ]

        p_list = []
        self._group_ids_by_parameter = []
        istate = []
        self._offsets = []
        self._dir1s = []
        self._dir2s = []

        # loop over the groups, collecting initial parameters and states
        for igp, pnl_ids in enumerate(self._panel_ids_by_group):

            panel_centres_in_lab_frame = []
            for i in pnl_ids:
                pnl = detector[i]
                im_size = pnl.get_image_size_mm()
                cntr = (
                    matrix.col(pnl.get_origin())
                    + 0.5 * matrix.col(pnl.get_fast_axis()) * im_size[0]
                    + 0.5 * matrix.col(pnl.get_slow_axis()) * im_size[1]
                )
                panel_centres_in_lab_frame.append(cntr)

            # get some vectors we need from the group
            go = matrix.col(self._groups[igp].get_origin())
            d1 = matrix.col(self._groups[igp].get_fast_axis())
            d2 = matrix.col(self._groups[igp].get_slow_axis())
            dn = matrix.col(self._groups[igp].get_normal())

            # we choose the dorg vector for this group to terminate on the group's
            # frame, at a point that we consider close to the centre of the group of
            # panels. This point is defined by taking the 3D centroid of the panel
            # centres then projecting that point onto the group frame.
            centroid = reduce(lambda a, b: a + b, panel_centres_in_lab_frame) / len(
                panel_centres_in_lab_frame
            )
            try:
                gp_centroid = matrix.col(
                    self._groups[igp].get_bidirectional_ray_intersection(centroid)
                )
                dorg = go + gp_centroid[0] * d1 + gp_centroid[1] * d2
            except RuntimeError:  # workaround for a group frame that passes through
                # the origin
                dorg = matrix.col((0.0, 0.0, 0.0))

            # The offset between the end of the dorg vector and
            # each Panel origin is a coordinate matrix with elements in the basis d1,
            # d2, dn. We need also each Panel's plane directions dir1 and dir2 in
            # terms of d1, d2 and dn.
            offsets, dir1s, dir2s = [], [], []
            for p in [detector[i] for i in pnl_ids]:
                offset = matrix.col(p.get_origin()) - dorg
                offsets.append(
                    matrix.col((offset.dot(d1), offset.dot(d2), offset.dot(dn)))
                )
                dir1 = matrix.col(p.get_fast_axis())
                dir1_new_basis = matrix.col((dir1.dot(d1), dir1.dot(d2), dir1.dot(dn)))
                dir1s.append(dir1_new_basis)
                dir2 = matrix.col(p.get_slow_axis())
                dir2_new_basis = matrix.col((dir2.dot(d1), dir2.dot(d2), dir2.dot(dn)))
                dir2s.append(dir2_new_basis)

            # The offsets and directions in the d1, d2, dn basis are fixed
            # quantities, not dependent on parameter values. Keep these as separate
            # sub-lists for each group
            self._offsets.append(offsets)
            self._dir1s.append(dir1s)
            self._dir2s.append(dir2s)

            # Set up the initial state for this group. This is the basis d1, d2, dn,
            # plus the offset locating the origin of the initial group frame
            gp_offset = go - dorg  # lab frame basis
            gp_offset = matrix.col(
                (gp_offset.dot(d1), gp_offset.dot(d2), gp_offset.dot(dn))
            )  # d1,d2,dn basis
            istate.append({"d1": d1, "d2": d2, "dn": dn, "gp_offset": gp_offset})

            # set up the parameters.
            # distance from lab origin to ref_panel plane along its normal,
            # in initial orientation
            distance = self._groups[igp].get_directed_distance()
            dist = Parameter(distance, dn, "length (mm)", f"Group{igp + 1}Dist")

            # shift in the detector model plane to locate dorg, in initial
            # orientation
            shift = dorg - dn * distance
            shift1 = Parameter(
                shift.dot(d1), d1, "length (mm)", f"Group{igp + 1}Shift1"
            )
            shift2 = Parameter(
                shift.dot(d2), d2, "length (mm)", f"Group{igp + 1}Shift2"
            )

            # rotations of the plane through its origin about:
            # 1) axis normal to initial orientation
            # 2) d1 axis of initial orientation
            # 3) d2 axis of initial orientation
            tau1 = Parameter(0, dn, "angle (mrad)", f"Group{igp + 1}Tau1")
            tau2 = Parameter(0, d1, "angle (mrad)", f"Group{igp + 1}Tau2")
            tau3 = Parameter(0, d2, "angle (mrad)", f"Group{igp + 1}Tau3")

            # extend the parameter list with those pertaining to this group
            p_list.extend([dist, shift1, shift2, tau1, tau2, tau3])
            self._group_ids_by_parameter.extend([igp] * 6)

        # set up the base class
        ModelParameterisation.__init__(
            self,
            detector,
            istate,
            p_list,
            experiment_ids=experiment_ids,
            is_multi_state=True,
        )

        # call compose to calculate all the derivatives
        self.compose()

    def get_panel_ids_by_group(self):
        """Return the panel IDs for each panel group of the detector.

        Provides access to a result calculated once during initialisation of the
        class, for use during parameterisation auto reduction tasks.

        Returns:
            A list over the panel groups. Each element of the list is itself a list,
                containing the panel IDs for that group.
        """
        return self._panel_ids_by_group

    def get_param_panel_groups(self):
        """Return the panel group ID for each parameter of the parameterisation.

        Provides access to a result calculated once during initialisation of the
        class, for use during parameterisation auto reduction tasks.

        Returns:
            A list over the parameters. Each element of the list contains the panel
                group ID of the panel group affected by that parameter.
        """
        return self._group_ids_by_parameter

    def compose(self):

        # reset the list that holds derivatives
        for i in range(len(self._model)):
            self._multi_state_derivatives[i] = [None] * len(self._dstate_dp)

        # loop over groups of panels collecting derivatives of the state wrt
        # parameters
        param = iter(self._param)
        for igp, pnl_ids in enumerate(self._panel_ids_by_group):

            # extract parameters from the internal list
            dist = next(param)
            shift1 = next(param)
            shift2 = next(param)
            tau1 = next(param)
            tau2 = next(param)
            tau3 = next(param)

            param_vals = flex.double(
                (
                    dist.value,
                    shift1.value,
                    shift2.value,
                    tau1.value,
                    tau2.value,
                    tau3.value,
                )
            )
            param_axes = flex.vec3_double(
                (dist.axis, shift1.axis, shift2.axis, tau1.axis, tau2.axis, tau3.axis)
            )

            offsets = self._offsets[igp]
            dir1s = self._dir1s[igp]
            dir2s = self._dir2s[igp]

            # Get items from the initial state for the group of interest
            initial_state = self._initial_state[igp]
            id1 = initial_state["d1"]
            id2 = initial_state["d2"]
            idn = initial_state["dn"]
            igp_offset = initial_state["gp_offset"]

            # Compose the new state using the helper class for calculations
            pgc = PanelGroupCompose(id1, id2, idn, igp_offset, param_vals, param_axes)

            # assign back to the group frame
            self._groups[igp].set_frame(pgc.d1(), pgc.d2(), pgc.origin())

            # Loop over attached Panel matrices, using the helper class to calculate
            # derivatives of the d matrix in each case and store them.
            i = igp * 6
            for (panel_id, offset, dir1_new_basis, dir2_new_basis) in zip(
                pnl_ids, offsets, dir1s, dir2s
            ):

                self._multi_state_derivatives[panel_id][
                    i : (i + 6)
                ] = pgc.derivatives_for_panel(offset, dir1_new_basis, dir2_new_basis)



from __future__ import annotations

import math
import pickle
import random


def test_getters():
    from dials.algorithms.profile_model.modeller import GridSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    depth = scan_range[1] - scan_range[0]
    nx = 10
    ny = 10
    nz = 2
    sampler = GridSampler((width, height), scan_range, (nx, ny, nz))
    image_size = sampler.image_size()
    scan_range = sampler.scan_range()
    grid_size = sampler.grid_size()
    step_size = sampler.step_size()
    size = len(sampler)

    assert width == image_size[0]
    assert height == image_size[1]
    assert scan_range[0] == scan_range[0]
    assert scan_range[1] == scan_range[1]
    assert nx == grid_size[0]
    assert ny == grid_size[1]
    assert nz == grid_size[2]
    assert step_size[0] == width / nx
    assert step_size[1] == height / ny
    assert step_size[2] == depth / nz
    assert nx * ny * nz == size


def test_indexing():
    from dials.algorithms.profile_model.modeller import GridSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nx = 10
    ny = 10
    nz = 2
    sampler = GridSampler((width, height), scan_range, (nx, ny, nz))
    xstep, ystep, zstep = sampler.step_size()
    xind = [list(range(nx))] * ny * nz
    yind = [[j] * nx for j in range(ny)] * nz
    zind = [[k] * nx * ny for k in range(nz)]
    xind = [i for j in xind for i in j]
    yind = [i for j in yind for i in j]
    zind = [i for j in zind for i in j]

    xp = [(x + 0.5) * xstep for x in xind]
    yp = [(y + 0.5) * ystep for y in yind]
    zp = [(z + 0.5) * zstep + scan_range[0] for z in zind]

    eps = 1e-10

    for x0, y0, z0, i in zip(xp, yp, zp, range(len(sampler))):
        x1, y1, z1 = sampler.coord(i)
        assert abs(x0 - x1) <= eps
        assert abs(y0 - y1) <= eps
        assert abs(z0 - z1) <= eps


def test_nearest():
    from dials.algorithms.profile_model.modeller import GridSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    depth = scan_range[1] - scan_range[0]
    nx = 10
    ny = 10
    nz = 2
    sampler = GridSampler((width, height), scan_range, (nx, ny, nz))

    for i in range(1000):
        x = random.uniform(0, 1000)
        y = random.uniform(0, 1000)
        z = random.uniform(*scan_range)
        i = int(math.floor(x / (width / nx)))
        j = int(math.floor(y / (height / ny)))
        k = int(math.floor((z - scan_range[0]) / (depth / nz)))
        if i >= nx:
            i = nx - 1
        if j >= ny:
            j = ny - 1
        if k >= nz:
            k = nz - 1
        index0 = i + j * nx + k * nx * ny
        index1 = sampler.nearest(0, (x, y, z))
        assert index0 == index1


def test_nearest_n():
    from dials.algorithms.profile_model.modeller import GridSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    depth = scan_range[1] - scan_range[0]
    nx = 10
    ny = 10
    nz = 2
    sampler = GridSampler((width, height), scan_range, (nx, ny, nz))

    for i in range(1000):
        x = random.uniform(0, 1000)
        y = random.uniform(0, 1000)
        z = random.uniform(*scan_range)
        i = int(math.floor(x * nx / width))
        j = int(math.floor(y * ny / height))
        k = int(math.floor((z - scan_range[0]) * nz / depth))
        if i >= nx:
            i = nx - 1
        if j >= ny:
            j = ny - 1
        if k >= nz:
            k = nz - 1
        index0 = i + j * nx + k * nx * ny
        index1 = sampler.nearest_n(0, (x, y, z))
        assert len(set(index1)) == len(index1)
        assert index0 == index1[-1]
        for ind in index1:
            ii = ind % nx
            jk = ind // nx
            jj = jk % ny
            kk = jk // ny
            assert abs(ii - i) <= 1
            assert abs(jj - j) <= 1
            assert abs(kk - k) <= 1


def test_weights():
    from scitbx import matrix

    from dials.algorithms.profile_model.modeller import GridSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nx = 10
    ny = 10
    nz = 2
    sampler = GridSampler((width, height), scan_range, (nx, ny, nz))

    # Check the weight at the coord in 1.0
    eps = 1e-7
    for i in range(len(sampler)):
        coord = sampler.coord(i)
        weight = sampler.weight(i, 0, coord)
        assert abs(weight - 1.0) < eps

    # Ensure we get the expected weight at the next grid point at half way
    # between grid points
    expected = math.exp(-4.0 * math.log(2.0))
    for k in range(nz):
        for j in range(ny):
            for i in range(nx):
                l1 = (i + 0) + ((j + 0) + (k + 0) * ny) * nx
                l2 = (i + 1) + ((j + 0) + (k + 0) * ny) * nx
                l3 = (i - 1) + ((j + 0) + (k + 0) * ny) * nx
                l4 = (i + 0) + ((j + 1) + (k + 0) * ny) * nx
                l5 = (i + 0) + ((j - 1) + (k + 0) * ny) * nx
                l6 = (i + 0) + ((j + 0) + (k + 1) * ny) * nx
                l7 = (i + 0) + ((j + 0) + (k - 1) * ny) * nx
                coord1 = matrix.col(sampler.coord(l1))
                if i < nx - 1:
                    coord = matrix.col(sampler.coord(l2))
                    weight = sampler.weight(l1, 0, coord)
                    assert abs(weight - expected) < eps
                    weight = sampler.weight(l1, 0, (coord + coord1) / 2.0)
                    assert abs(weight - 0.5) < eps
                if i > 0:
                    coord = matrix.col(sampler.coord(l3))
                    weight = sampler.weight(l1, 0, coord)
                    assert abs(weight - expected) < eps
                    weight = sampler.weight(l1, 0, (coord1 + coord) / 2.0)
                    assert abs(weight - 0.5) < eps
                if j < ny - 1:
                    coord = matrix.col(sampler.coord(l4))
                    weight = sampler.weight(l1, 0, coord)
                    assert abs(weight - expected) < eps
                    weight = sampler.weight(l1, 0, (coord + coord1) / 2.0)
                    assert abs(weight - 0.5) < eps
                if j > 0:
                    coord = matrix.col(sampler.coord(l5))
                    weight = sampler.weight(l1, 0, coord)
                    assert abs(weight - expected) < eps
                    weight = sampler.weight(l1, 0, (coord1 + coord) / 2.0)
                    assert abs(weight - 0.5) < eps
                if k < nz - 1:
                    coord = matrix.col(sampler.coord(l6))
                    weight = sampler.weight(l1, 0, coord)
                    assert abs(weight - expected) < eps
                    weight = sampler.weight(l1, 0, (coord + coord1) / 2.0)
                    assert abs(weight - 0.5) < eps
                if k > 0:
                    coord = matrix.col(sampler.coord(l7))
                    weight = sampler.weight(l1, 0, coord)
                    assert abs(weight - expected) < eps
                    weight = sampler.weight(l1, 0, (coord1 + coord) / 2.0)
                    assert abs(weight - 0.5) < eps


def test_self_consistent():
    from dials.algorithms.profile_model.modeller import GridSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nx = 10
    ny = 10
    nz = 2
    sampler = GridSampler((width, height), scan_range, (nx, ny, nz))

    for i in range(len(sampler)):
        coord = sampler.coord(i)
        index = sampler.nearest(0, coord)
        assert index == i


def test_pickle():
    from dials.algorithms.profile_model.modeller import GridSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nx = 10
    ny = 10
    nz = 2
    sampler = GridSampler((width, height), scan_range, (nx, ny, nz))

    sampler2 = pickle.loads(pickle.dumps(sampler))

    assert sampler.image_size() == sampler2.image_size()
    assert sampler.grid_size() == sampler2.grid_size()



"""
Optimise the combination of profile and summation intensity values.
"""

from __future__ import annotations

import logging

import boost_adaptbx.boost.python
from cctbx import crystal, miller

from dials.algorithms.scaling.scaling_utilities import DialsMergingStatisticsError
from dials.array_family import flex
from dials.util import tabulate

miller_ext = boost_adaptbx.boost.python.import_ext("cctbx_miller_ext")
logger = logging.getLogger("dials")


def fast_merging_stats(array):
    """
    Quickly calculate required merging stats for intensity combination.

    This is a cut-down version of iobtx.merging_statistics.merging_stats.
    """
    assert array.sigmas() is not None
    positive_sel = array.sigmas() > 0
    i_over_sigma_sel = (array.data() / array.sigmas()) > 1.0
    array = array.select(positive_sel & i_over_sigma_sel)
    if not array.size():
        return -1.0, -1.0
    array = array.sort("packed_indices")
    merge_ext = miller_ext.merge_equivalents_obs(
        array.indices(), array.data(), array.sigmas(), use_internal_variance=True
    )
    r_meas = merge_ext.r_meas
    cc_one_half = miller.compute_cc_one_half(unmerged=array, return_n_refl=False)
    return r_meas, cc_one_half


def map_indices_to_asu(miller_indices, space_group):
    """Map the indices to the asymmetric unit."""
    crystal_symmetry = crystal.symmetry(space_group=space_group)
    miller_set = miller.set(
        crystal_symmetry=crystal_symmetry, indices=miller_indices, anomalous_flag=False
    )
    miller_set_in_asu = miller_set.map_to_asu()
    return miller_set_in_asu.indices()


def _make_reflection_table_from_scaler(scaler):
    """Copy across required columns and filter data."""
    reflections = flex.reflection_table()
    required_cols = [
        "intensity.prf.value",
        "intensity.prf.variance",
        "intensity.sum.value",
        "intensity.sum.variance",
        "prescaling_correction",
        "inverse_scale_factor",
        "miller_index",
    ]
    optional_cols = ["partiality"]
    for col in required_cols:
        reflections[col] = scaler.reflection_table[col]
    for col in optional_cols:
        if col in scaler.reflection_table:
            reflections[col] = scaler.reflection_table[col]
    # now select good data
    sel = _get_filter_selection(scaler.reflection_table)
    suitable_isel = scaler.suitable_refl_for_scaling_sel.iselection()
    outlier_isel = suitable_isel.select(scaler.outliers)
    free_set_isel = suitable_isel.select(scaler.free_set_selection)
    not_outliers_or_free = flex.bool(reflections.size(), True)
    not_outliers_or_free.set_selected(outlier_isel, False)
    not_outliers_or_free.set_selected(free_set_isel, False)
    reflections = reflections.select(sel & not_outliers_or_free)
    reflections["miller_index"] = map_indices_to_asu(
        reflections["miller_index"], scaler.experiment.crystal.get_space_group()
    )
    logger.debug("Reflection table size for combining: %s", reflections.size())
    return reflections


def _determine_Imids(combiner, raw_intensities):
    if not combiner.Imids:
        avg = max(10, flex.mean(raw_intensities))
        Imid = flex.max(raw_intensities) / 10.0
        Imid_list = [0, 1, avg, Imid]
        while Imid > avg:
            Imid /= 10.0
            Imid_list.append(Imid)
        combiner.Imids = Imid_list


class SingleDatasetIntensityCombiner:
    """
    Class to combine profile and summation intensities for a single dataset.
    """

    def __init__(self, scaler, use_Imid=None):
        self.scaler = scaler
        self.experiment = scaler.experiment
        if "intensity.prf.value" not in scaler.reflection_table:
            self.max_key = 1
            logger.info(
                "No profile intensities found, skipping profile/summation intensity combination."
            )
            return
        if use_Imid is not None:
            self.max_key = use_Imid
        else:
            self.Imids = scaler.params.reflection_selection.combine.Imid
            self.dataset = _make_reflection_table_from_scaler(self.scaler)
            if "partiality" in self.dataset:
                raw_intensities = (
                    self.dataset["intensity.sum.value"].as_double()
                    / self.dataset["partiality"]
                )
            else:
                raw_intensities = self.dataset["intensity.sum.value"].as_double()
            logger.debug("length of raw intensity array: %s", raw_intensities.size())
            _determine_Imids(self, raw_intensities)
            header = ["Combination", "CC1/2", "Rmeas"]
            rows, results = self._test_Imid_combinations()
            logger.info(tabulate(rows, header))

            self.max_key = min(results, key=results.get)
            while results[self.max_key] < 0:
                del results[self.max_key]
                if results:
                    self.max_key = min(results, key=results.get)
                else:
                    self.max_key = -1
                    break
            if self.max_key == 0:
                logger.info("Profile intensities determined to be best for scaling. \n")
            elif self.max_key == 1:
                logger.info(
                    "Summation intensities determined to be best for scaling. \n"
                )
            elif self.max_key == -1:
                logger.info("No good statistics found, using profile intensities. \n")
                self.max_key = 0
            else:
                logger.info(
                    "Combined intensities with Imid = %.2f determined to be best for scaling. \n",
                    self.max_key,
                )

    def calculate_suitable_combined_intensities(self):
        """Combine the 'suitable for scaling' intensities in the scaler."""
        return _calculate_suitable_combined_intensities(self.scaler, self.max_key)

    def _test_Imid_combinations(self):
        """Test the different combinations, returning the rows and results dict."""
        rows = []
        results = {}

        for Imid in self.Imids:
            Int, Var = _get_Is_from_Imidval(self.dataset, Imid)
            miller_set = miller.set(
                crystal_symmetry=self.experiment.crystal.get_crystal_symmetry(
                    assert_is_compatible_unit_cell=False
                ),
                indices=self.dataset["miller_index"],
                anomalous_flag=False,
            )
            i_obs = miller.array(
                miller_set,
                data=(
                    Int
                    * self.dataset["prescaling_correction"]
                    / self.dataset["inverse_scale_factor"]
                ),
            )
            i_obs.set_observation_type_xray_intensity()
            i_obs.set_sigmas(
                flex.sqrt(Var)
                * self.dataset["prescaling_correction"]
                / self.dataset["inverse_scale_factor"]
            )
            try:
                rmeas, cchalf = fast_merging_stats(array=i_obs)
                logger.debug("Imid: %s, Rmeas %s, cchalf %s", Imid, rmeas, cchalf)
            except RuntimeError:
                raise DialsMergingStatisticsError(
                    "Unable to merge for intensity combination"
                )

            # record the results
            results[Imid] = rmeas
            res_str = {0: "prf only", 1: "sum only"}
            if Imid not in res_str:
                res_str[Imid] = "Imid = " + str(round(Imid, 2))
            rows.append([res_str[Imid], str(round(cchalf, 5)), str(round(rmeas, 5))])

        return rows, results


def combine_intensities(reflections, Imid):
    """Take unscaled data, and apply intensity combination with a given Imid."""
    if "intensity.prf.value" in reflections:
        Ipr = reflections["intensity.prf.value"]
        Vpr = reflections["intensity.prf.variance"]
    assert "intensity.sum.value" in reflections
    assert "prescaling_correction" in reflections

    conv = reflections["prescaling_correction"]
    Isum = reflections["intensity.sum.value"]
    Vsum = reflections["intensity.sum.variance"]

    not_prf = ~reflections.get_flags(reflections.flags.integrated_prf)
    not_sum = ~reflections.get_flags(reflections.flags.integrated_sum)
    both = reflections.get_flags(reflections.flags.integrated, all=True)

    if "partiality" in reflections:
        inv_p = _determine_inverse_partiality(reflections)
        sum_conv = conv * inv_p
    else:
        sum_conv = conv

    if Imid == 1:  # i.e. sum is best, so use sum if exists, else prf
        intensity = Isum * sum_conv
        variance = Vsum * sum_conv * sum_conv
        # get not summation successful
        if "intensity.prf.value" in reflections:
            intensity.set_selected(not_sum.iselection(), (Ipr * conv).select(not_sum))
            variance.set_selected(
                not_sum.iselection(), (Vpr * conv * conv).select(not_sum)
            )
    else:
        # first set as prf
        intensity = Ipr * conv
        variance = Vpr * conv * conv
        # set those not prf successful
        intensity.set_selected(not_prf.iselection(), (Isum * sum_conv).select(not_prf))
        variance.set_selected(
            not_prf.iselection(), (Vsum * sum_conv * sum_conv).select(not_prf)
        )
        if Imid == 0:  # done all we need to do.
            pass
        else:
            # calculate combined intensities, but only set for those where both prf and sum good
            if "partiality" in reflections:
                Int, Var = _calculate_combined_raw_intensities(
                    Ipr, Isum * inv_p, Vpr, Vsum * inv_p * inv_p, Imid
                )
            else:
                Int, Var = _calculate_combined_raw_intensities(
                    Ipr, Isum, Vpr, Vsum, Imid
                )
            intensity.set_selected(both.iselection(), (Int * conv).select(both))
            variance.set_selected(both.iselection(), (Var * conv * conv).select(both))

    return intensity, variance


def _calculate_suitable_combined_intensities(scaler, max_key):
    reflections = scaler.reflection_table.select(scaler.suitable_refl_for_scaling_sel)
    return combine_intensities(reflections, max_key)


class MultiDatasetIntensityCombiner:
    """
    Class to combine profile and summation intensities for multiple datasets.
    """

    def __init__(self, multiscaler):
        self.active_scalers = multiscaler.active_scalers
        self.Imids = multiscaler.params.reflection_selection.combine.Imid
        # first copy across relevant data that's needed
        self.good_datasets = []
        for i, scaler in enumerate(self.active_scalers):
            if "intensity.prf.value" in scaler.reflection_table:
                self.good_datasets.append(i)
        if not self.good_datasets:
            self.max_key = 1
            logger.info(
                "No profile intensities found, skipping profile/summation intensity combination."
            )
            return
        self.datasets = [
            _make_reflection_table_from_scaler(self.active_scalers[i])
            for i in self.good_datasets
        ]
        raw_intensities = self._get_raw_intensity_array()
        logger.debug("length of raw intensity array: %s", raw_intensities.size())
        _determine_Imids(self, raw_intensities)

        header = ["Combination", "CC1/2", "Rmeas"]
        rows, results = self._test_Imid_combinations()
        logger.info(tabulate(rows, header))

        self.max_key = min(results, key=results.get)
        while results[self.max_key] < 0:
            del results[self.max_key]
            if results:
                self.max_key = min(results, key=results.get)
            else:
                self.max_key = -1
                break
        if self.max_key == 0:
            logger.info("Profile intensities determined to be best for scaling. \n")
        elif self.max_key == 1:
            logger.info("Summation intensities determined to be best for scaling. \n")
        elif self.max_key == -1:
            logger.info("No good statistics found, using profile intensities. \n")
            self.max_key = 0
        else:
            logger.info(
                "Combined intensities with Imid = %.2f determined to be best for scaling. \n",
                self.max_key,
            )

    def calculate_suitable_combined_intensities(self, dataset):
        """Combine the 'suitable for scaling' intensities in the scaler."""
        if dataset not in self.good_datasets:
            return _calculate_suitable_combined_intensities(
                self.active_scalers[dataset], 1
            )
        return _calculate_suitable_combined_intensities(
            self.active_scalers[dataset], self.max_key
        )

    def _get_raw_intensity_array(self):
        intensities = flex.double()
        for dataset in self.datasets:
            if "partiality" in dataset:
                intensities.extend(
                    dataset["intensity.sum.value"].as_double() / dataset["partiality"]
                )
            else:
                intensities.extend(dataset["intensity.sum.value"].as_double())
        return intensities

    def _test_Imid_combinations(self):
        rows = []
        results = {}
        for Imid in self.Imids:
            combined_intensities = flex.double([])
            combined_sigmas = flex.double([])
            combined_scales = flex.double([])
            combined_indices = flex.miller_index([])
            for dataset in self.datasets:
                Int, Var = _get_Is_from_Imidval(dataset, Imid)
                sigma = flex.sqrt(Var) * dataset["prescaling_correction"]
                combined_intensities.extend(Int * dataset["prescaling_correction"])
                combined_sigmas.extend(sigma)
                combined_scales.extend(dataset["inverse_scale_factor"])
                combined_indices.extend(dataset["miller_index"])
            # apply scale factor before determining merging stats
            miller_set = miller.set(
                crystal_symmetry=self.active_scalers[
                    0
                ].experiment.crystal.get_crystal_symmetry(),
                indices=combined_indices,
                anomalous_flag=False,
            )
            i_obs = miller.array(
                miller_set, data=combined_intensities / combined_scales
            )
            i_obs.set_observation_type_xray_intensity()
            i_obs.set_sigmas(combined_sigmas / combined_scales)
            try:
                rmeas, cchalf = fast_merging_stats(array=i_obs)
                logger.debug("Imid: %s, Rmeas %s, cchalf %s", Imid, rmeas, cchalf)
            except RuntimeError:
                raise DialsMergingStatisticsError(
                    "Unable to merge for intensity combination"
                )

            # record the results
            results[Imid] = rmeas
            res_str = {0: "prf only", 1: "sum only"}
            if Imid not in res_str:
                res_str[Imid] = "Imid = " + str(round(Imid, 2))
            rows.append([res_str[Imid], str(round(cchalf, 5)), str(round(rmeas, 5))])
        return rows, results


### Helper functions for combine_intensities


def _get_Is_from_Imidval(reflections, Imid):
    """Interpret the Imid value to extract and return the Icomb and Vcomb values."""
    if Imid == 0:  # special value to trigger prf
        Int = reflections["intensity.prf.value"]
        Var = reflections["intensity.prf.variance"]
    elif Imid == 1:  # special value to trigger sum
        if "partiality" in reflections:
            Int = reflections["intensity.sum.value"] / reflections["partiality"]
            Var = reflections["intensity.sum.variance"] / flex.pow2(
                reflections["partiality"]
            )
        else:
            Int = reflections["intensity.sum.value"]
            Var = reflections["intensity.sum.variance"]
    else:
        if "partiality" in reflections:
            Int, Var = _calculate_combined_raw_intensities(
                reflections["intensity.prf.value"],
                reflections["intensity.sum.value"] / reflections["partiality"],
                reflections["intensity.prf.variance"],
                reflections["intensity.sum.variance"]
                / flex.pow2(reflections["partiality"]),
                Imid,
            )
        else:
            Int, Var = _calculate_combined_raw_intensities(
                reflections["intensity.prf.value"],
                reflections["intensity.sum.value"],
                reflections["intensity.prf.variance"],
                reflections["intensity.sum.variance"],
                Imid,
            )
    return Int, Var


def _get_filter_selection(reflections):
    bad_sel = (
        reflections.get_flags(reflections.flags.bad_for_scaling, all=False)
        | (reflections["intensity.prf.variance"] <= 0)
        | (reflections["intensity.sum.variance"] <= 0)
        | (reflections["inverse_scale_factor"] <= 0)
    )
    integrated = reflections.get_flags(reflections.flags.integrated, all=True)
    good_sel = integrated & ~bad_sel
    if "partiality" in reflections:
        good_sel &= reflections["partiality"] > 0
    return good_sel


def _determine_inverse_partiality(reflections):
    inverse_partiality = flex.double(reflections.size(), 1.0)
    nonzero_partiality_sel = reflections["partiality"] > 0.0
    good_refl = reflections.select(nonzero_partiality_sel)
    inverse_partiality.set_selected(
        nonzero_partiality_sel.iselection(), 1.0 / good_refl["partiality"]
    )
    return inverse_partiality


def _calculate_combined_raw_intensities(Iprf, Isum, Vprf, Vsum, Imid):
    """Use partiality-corrected Isum, alongside Iprf to calculate
    combined raw intensities."""
    w = 1.0 / (1.0 + (Isum / Imid) ** 3)
    w.set_selected(Isum <= 0, 1.0)
    Icomb = (w * Iprf) + ((1.0 - w) * Isum)
    Vcomb = (w * Vprf) + ((1.0 - w) * Vsum)
    return Icomb, Vcomb



from __future__ import annotations

from os.path import basename, splitext

from dxtbx.model.experiment_list import ExperimentList

import dials.util
from dials.algorithms.refinement.refinement_helpers import calculate_frame_numbers
from dials.array_family import flex
from dials.util import Sorry
from dials.util.multi_dataset_handling import generate_experiment_identifiers
from dials.util.slice import slice_experiments, slice_reflections

help_message = """

Slice a sequence to produce a smaller sequence within the bounds of the original. If
experiments or experiments are provided, modify the scan objects within these. If
reflections are provided, remove reflections outside the provided image ranges.
Each image_range parameter refers to a single experiment ID, counting up from
zero. Any reflections with experiment ID not matched by a image_range parameter
are removed.

Examples::

  dials.slice_sequence models.expt observations.refl "image_range=1 20"

  dials.slice_sequence models.expt "image_range=1 20"

  # two experiments and reflections with IDs '0' and '1'
  dials.slice_sequence models.expt observations.refl \
    "image_range=1 20" "image_range=5 30"
"""

from libtbx.phil import parse

phil_scope = parse(
    """

  output {

    reflections_filename = None
      .type = str
      .help = "The filename for output reflections sliced to remove those"
              "outside the reduced image range. By default generated"
              "automatically from the input name"

    experiments_filename = None
      .type = str
      .help = "The filename for the output experiments with sliced scans.
               By default generated automatically from the input name"

  }

  image_range = None
    .help = "Range in images to slice a sequence. The number of arguments"
            "must be a factor of two. Each pair of arguments gives a range"
            "that follows C conventions (e.g. j0 <= j < j1) when slicing the"
            "reflections by observed centroid."
    .type = ints(size=2)
    .multiple = True

  block_size = None
    .type = float
    .help = "Overrides image_range if present. This option splits each sequence"
            "into the nearest integer number of equal size blocks close to"
            "block_size degrees in width"
"""
)


def calculate_block_ranges(scan, block_size):
    """

    :param scans
    :type a scan object
    :param block_size:
    :type block_size: target block size in degrees"""

    image_ranges = []
    nimages = scan.get_num_images()
    osc_range = scan.get_oscillation_range(deg=True)
    osc_width = abs(osc_range[1] - osc_range[0])
    nblocks = max(int(round(osc_width / block_size)), 1)
    nblocks = min(nblocks, nimages)
    # equal sized blocks except the last one that may contain extra images
    # to make up the remainder
    nimages_per_block = [nimages // nblocks] * (nblocks - 1) + [
        nimages // nblocks + nimages % nblocks
    ]
    start = scan.get_image_range()[0]
    for nim in nimages_per_block:
        image_ranges.append((start, start + nim - 1))
        start += nim

    return image_ranges


class Script:
    """A class for running the script."""

    def __init__(self):
        """Initialise the script."""
        from dials.util.options import ArgumentParser

        usage = (
            "usage: dials.slice_sequence [options] [param.phil] "
            "models.expt observations.refl"
        )

        # Create the parser
        self.parser = ArgumentParser(
            usage=usage,
            phil=phil_scope,
            read_reflections=True,
            read_experiments=True,
            check_format=False,
            epilog=help_message,
        )

    def run(self, args=None):
        """Execute the script."""

        from dials.util.options import reflections_and_experiments_from_files

        # Parse the command line
        params, options = self.parser.parse_args(args, show_diff_phil=True)
        reflections, experiments = reflections_and_experiments_from_files(
            params.input.reflections, params.input.experiments
        )

        # Try to load the models and data
        slice_exps = len(experiments) > 0
        slice_refs = len(reflections) > 0

        # Catch case of nothing to do
        if not slice_exps and not slice_refs:
            print("No suitable input provided")
            self.parser.print_help()
            return

        if reflections:
            if len(reflections) > 1:
                raise Sorry("Only one reflections list can be imported at present")
            reflections = reflections[0]

            # calculate frame numbers if needed
            if experiments:
                reflections = calculate_frame_numbers(reflections, experiments)

            # if we still don't have the right column give up
            if "xyzobs.px.value" not in reflections:
                raise Sorry(
                    "These reflections do not have frame numbers set, and "
                    "there are no experiments provided to calculate these."
                )

        # set trivial case where no scan range is provided at all
        if not params.image_range:
            params.image_range = [None]

        # check if slicing into blocks
        if params.block_size is not None:
            if not slice_exps:
                raise Sorry(
                    "For slicing into blocks, an experiment file must be provided"
                )

            if len(experiments) > 1:
                raise Sorry("For slicing into blocks please provide a single scan only")
            scan = experiments[0].scan

            # Having extracted the scan, calculate the blocks
            params.image_range = calculate_block_ranges(scan, params.block_size)

            # Do the slicing then recombine
            sliced = [
                slice_experiments(experiments, [sr])[0] for sr in params.image_range
            ]
            generate_experiment_identifiers(sliced)
            sliced_experiments = ExperimentList(sliced)

            # slice reflections if present
            if slice_refs:
                sliced = [
                    slice_reflections(reflections, [sr]) for sr in params.image_range
                ]
                sliced_reflections = flex.reflection_table()
                identifiers = sliced_experiments.identifiers()
                # resetting experiment identifiers
                for i, rt in enumerate(sliced):
                    for k in rt.experiment_identifiers().keys():
                        del rt.experiment_identifiers()[k]
                    rt["id"] = flex.int(rt.size(), i)  # set id
                    rt.experiment_identifiers()[i] = identifiers[i]
                    sliced_reflections.extend(rt)

        else:
            # slice each dataset into the requested subset
            if slice_exps:
                sliced_experiments = slice_experiments(experiments, params.image_range)
            if slice_refs:
                sliced_reflections = slice_reflections(reflections, params.image_range)

        # Save sliced experiments
        if slice_exps:
            output_experiments_filename = params.output.experiments_filename
            if output_experiments_filename is None:
                # take first filename as template
                bname = basename(params.input.experiments[0].filename)
                bname = splitext(bname)[0]
                if not bname:
                    bname = "experiments"
                if len(params.image_range) == 1 and params.image_range[0] is not None:
                    ext = "_{}_{}.expt".format(*params.image_range[0])
                else:
                    ext = "_sliced.expt"
                output_experiments_filename = bname + ext
            print(f"Saving sliced experiments to {output_experiments_filename}")

            sliced_experiments.as_file(output_experiments_filename)

        # Save sliced reflections
        if slice_refs:
            output_reflections_filename = params.output.reflections_filename
            if output_reflections_filename is None:
                # take first filename as template
                bname = basename(params.input.reflections[0].filename)
                bname = splitext(bname)[0]
                if not bname:
                    bname = "reflections"
                if len(params.image_range) == 1 and params.image_range[0] is not None:
                    ext = "_{}_{}.refl".format(*params.image_range[0])
                else:
                    ext = "_sliced.refl"
                output_reflections_filename = bname + ext

            print(f"Saving sliced reflections to {output_reflections_filename}")
            sliced_reflections.as_file(output_reflections_filename)

        return


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()
